from datetime import datetime, timedelta
from ipaddress import IPv4Address
from typing import Tuple
from uuid import UUID

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509 import NameAttribute, NameOID


def generate_self_signed_certificate(server_uuid: UUID, address: str) -> Tuple[bytes, bytes]:
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

    issuer = subject = get_sila_consortium_x509_name()

    today = datetime.today()
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(private_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(today - timedelta(hours=12))
        .not_valid_after(today + timedelta(days=364))
        .add_extension(
            x509.BasicConstraints(ca=False, path_length=None),
            critical=True,
        )
        .add_extension(
            x509.ExtendedKeyUsage([x509.oid.ExtendedKeyUsageOID.SERVER_AUTH, x509.oid.ExtendedKeyUsageOID.CLIENT_AUTH]),
            critical=False,
        )
        .add_extension(
            x509.KeyUsage(
                digital_signature=True,
                key_encipherment=True,
                key_cert_sign=True,
                crl_sign=False,
                key_agreement=False,
                encipher_only=False,
                decipher_only=False,
                content_commitment=False,
                data_encipherment=False,
            ),
            critical=True,
        )
        .add_extension(
            x509.SubjectAlternativeName([x509.IPAddress(IPv4Address(address))]),
            critical=False,
        )
        .add_extension(
            x509.UnrecognizedExtension(x509.ObjectIdentifier("1.3.6.1.4.1.58583"), str(server_uuid).encode("ascii")),
            critical=False,
        )
        .add_extension(
            x509.SubjectKeyIdentifier.from_public_key(private_key.public_key()),
            critical=False,
        )
        .sign(private_key, hashes.SHA256())
    )

    private_key_bytes = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )
    cert_bytes = cert.public_bytes(serialization.Encoding.PEM)

    return private_key_bytes, cert_bytes


def get_sila_consortium_x509_name() -> x509.Name:
    return x509.Name(
        [
            NameAttribute(NameOID.COUNTRY_NAME, "CH"),  # Switzerland
            NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "SG"),  # Canton of St. Gallen
            NameAttribute(NameOID.LOCALITY_NAME, "Rapperswil-Jona"),
            NameAttribute(NameOID.ORGANIZATION_NAME, "Association Consortium Standardization in Lab Automation (SiLA)"),
            NameAttribute(NameOID.COMMON_NAME, "SiLA2"),
        ]
    )



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from sila2.framework.abc.sila_error import SilaError

if TYPE_CHECKING:
    from sila2.client.sila_client import SilaClient
    from sila2.pb2_stubs.SiLAFramework_pb2 import FrameworkError as SilaFrameworkError
    from sila2.pb2_stubs.SiLAFramework_pb2 import SiLAError


class FrameworkError(SilaError):
    """
    Base class for all SiLA Framework Errors. Don't use this class directly. Instead, use one of its subclasses:

    - :py:class:`sila2.framework.InvalidCommandExecutionUUID`
    - :py:class:`sila2.framework.CommandExecutionNotFinished`
    - :py:class:`sila2.framework.InvalidMetadata`
    - :py:class:`sila2.framework.NoMetadataAllowed`
    - :py:class:`sila2.framework.CommandExecutionNotAccepted`
    """

    error_type: FrameworkErrorType
    message: str

    def __init__(self, error_type: FrameworkErrorType, message: str):
        self.error_type = error_type
        self.message = message

    def to_message(self) -> SiLAError:
        return self._pb2_module.SiLAError(
            frameworkError=self._pb2_module.FrameworkError(
                errorType=getattr(self._pb2_module.FrameworkError, self.error_type.name),
                message=self.message,
            )
        )

    @classmethod
    def from_message(cls, message: SilaFrameworkError, client: SilaClient) -> FrameworkError:
        if message.errorType == message.COMMAND_EXECUTION_NOT_ACCEPTED:
            from sila2.framework.errors.command_execution_not_accepted import CommandExecutionNotAccepted

            return CommandExecutionNotAccepted(message.message)
        if message.errorType == message.INVALID_COMMAND_EXECUTION_UUID:
            from sila2.framework.errors.invalid_command_execution_uuid import InvalidCommandExecutionUUID

            return InvalidCommandExecutionUUID(message.message)
        if message.errorType == message.COMMAND_EXECUTION_NOT_FINISHED:
            from sila2.framework.errors.command_execution_not_finished import CommandExecutionNotFinished

            return CommandExecutionNotFinished(message.message)
        if message.errorType == message.INVALID_METADATA:
            from sila2.framework.errors.invalid_metadata import InvalidMetadata

            return InvalidMetadata(message.message)
        if message.errorType == message.NO_METADATA_ALLOWED:
            from sila2.framework.errors.no_metadata_allowed import NoMetadataAllowed

            return NoMetadataAllowed(message.message)

        raise NotImplementedError(f"FrameworkError type not supported: {message.errorType}")  # should not happen


class FrameworkErrorType(Enum):
    COMMAND_EXECUTION_NOT_ACCEPTED = 0
    INVALID_COMMAND_EXECUTION_UUID = 1
    COMMAND_EXECUTION_NOT_FINISHED = 2
    INVALID_METADATA = 3
    NO_METADATA_ALLOWED = 4



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional, Type

from google.protobuf.message import Message

from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.defined_execution_error_node import DefinedExecutionErrorNode
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.framework.utils import xpath_sila

if TYPE_CHECKING:
    from sila2.framework.feature import Feature


class Property(NamedDataNode):
    """Represents a property"""

    fully_qualified_identifier: FullyQualifiedIdentifier
    """Fully qualified property identifier"""
    parent_feature: Feature
    parameter_message_type: Type[Message]
    response_message_type: Type[Message]
    defined_execution_errors: List[DefinedExecutionErrorNode]

    def __init__(self, fdl_node, parent_feature: Feature):
        identifier: str = xpath_sila(fdl_node, "sila:Identifier/text()")[0]
        observable = xpath_sila(fdl_node, "sila:Observable/text()")[0] == "Yes"
        message_prefix = "Subscribe_" if observable else "Get_"
        super().__init__(
            fdl_node, parent_feature, getattr(parent_feature._pb2_module, f"{message_prefix}{identifier}_Responses")
        )
        self.parent_feature = parent_feature
        self.fully_qualified_identifier = FullyQualifiedIdentifier(
            f"{parent_feature.fully_qualified_identifier}/Property/{self._identifier}"
        )
        self.defined_execution_errors = [
            parent_feature.defined_execution_errors[name]
            for name in xpath_sila(fdl_node, "sila:DefinedExecutionErrors/sila:Identifier/text()")
        ]

    @staticmethod
    def from_fdl_node(fdl_node, parent_feature: Feature) -> Property:
        if xpath_sila(fdl_node, "sila:Observable/text() = 'No'"):
            from sila2.framework.property.unobservable_property import UnobservableProperty

            return UnobservableProperty(fdl_node, parent_feature)

        from sila2.framework.property.observable_property import ObservableProperty

        return ObservableProperty(fdl_node, parent_feature)

    def get_parameters_message(self) -> Message:
        return self.parameter_message_type()

    def to_native_type(self, msg: Message, toplevel_named_data_node: Optional[NamedDataNode] = None) -> Any:
        return self.data_type.to_native_type(
            getattr(msg, self._identifier), toplevel_named_data_node=toplevel_named_data_node
        )

    def to_message(self, value: Any, toplevel_named_data_node: Optional[NamedDataNode] = None) -> Message:
        return self.response_message_type(
            **{self._identifier: self.data_type.to_message(value, toplevel_named_data_node=toplevel_named_data_node)}
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, NamedTuple, Optional, Tuple, TypeVar, Union

from sila2.framework.abc.constraint import Constraint
from sila2.framework.utils import xpath_sila

if TYPE_CHECKING:
    from sila2.framework.abc.data_type import DataType
    from sila2.framework.feature import Feature


T = TypeVar("T", bound=Union[str, bytes])


class ContentType(Constraint[T]):
    type: str
    subtype: str
    parameters: Tuple[ContentTypeParameter, ...]

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        type: str,
        subtype: str,
        parameters: Optional[Iterable[ContentTypeParameter]] = None,
    ):
        self.type = type
        self.subtype = subtype
        self.parameters = () if parameters is None else tuple(parameters)

    def validate(self, value: T) -> bool:
        return True

    @property
    def media_type(self) -> str:
        media_type = f"{self.type}/{self.subtype}"
        for par in self.parameters:
            media_type += f"; {par.attribute}={par.value}"
        return media_type

    @classmethod
    def from_fdl_node(cls, fdl_node, parent_feature: Feature, base_type: DataType) -> ContentType:
        _type = xpath_sila(fdl_node, "sila:Type/text()")[0]
        subtype = xpath_sila(fdl_node, "sila:Subtype/text()")[0]

        parameters = []
        for par_node in xpath_sila(fdl_node, "sila:Parameters/sila:Parameter"):
            attribute = xpath_sila(par_node, "sila:Attribute/text()")[0]
            value = xpath_sila(par_node, "sila:Value/text()")[0]
            parameters.append(ContentTypeParameter(attribute, value))

        return cls(_type, subtype, parameters)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.media_type!r})"


class ContentTypeParameter(NamedTuple):
    attribute: str
    value: str



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict

from sila2.framework import FullyQualifiedIdentifier
from sila2.server import FeatureImplementationBase

from .authorizationproviderservice_types import Verify_Responses


class AuthorizationProviderServiceBase(FeatureImplementationBase, ABC):
    @abstractmethod
    def Verify(
        self,
        AccessToken: str,
        RequestedServer: str,
        RequestedFeature: str,
        *,
        metadata: Dict[FullyQualifiedIdentifier, Any],
    ) -> Verify_Responses:
        """
        Verifies that a given token is valid for the requested server.


        :param AccessToken: The token to be used along with accessing a Command or Property on a SiLA Server.

        :param RequestedServer: The server for which an authorization is requested.

        :param RequestedFeature: The feature that is requested to access.

        :param metadata: The SiLA Client Metadata attached to the call

        :return:

            - TokenLifetime: The lifetime of the provided access token as the maximum validity period after the last SiLA Server request.


        """
        pass



from __future__ import annotations

from typing import TYPE_CHECKING, List

from sila2.framework.abc.composite_message_mappable import CompositeMessageMappable
from sila2.framework.abc.named_node import NamedNode
from sila2.framework.command.parameter import Parameter
from sila2.framework.command.response import Response
from sila2.framework.defined_execution_error_node import DefinedExecutionErrorNode
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.framework.utils import xpath_sila

if TYPE_CHECKING:
    from sila2.framework.feature import Feature


class Command(NamedNode):
    """Represents a command"""

    fully_qualified_identifier: FullyQualifiedIdentifier
    """Fully qualified command identifier"""
    parent_feature: Feature
    parameters: CompositeMessageMappable[Parameter]
    responses: CompositeMessageMappable[Response]
    defined_execution_errors: List[DefinedExecutionErrorNode]

    def __init__(self, fdl_node, parent_feature: Feature):
        super().__init__(fdl_node)
        self.parent_feature = parent_feature
        self.fully_qualified_identifier = FullyQualifiedIdentifier(
            f"{parent_feature.fully_qualified_identifier}/Command/{self._identifier}"
        )
        self.parameters = CompositeMessageMappable(
            [Parameter(node, self) for node in xpath_sila(fdl_node, "sila:Parameter")],
            getattr(self.parent_feature._pb2_module, f"{self._identifier}_Parameters"),
        )
        self.responses = CompositeMessageMappable(
            [Response(node, self) for node in xpath_sila(fdl_node, "sila:Response")],
            getattr(self.parent_feature._pb2_module, f"{self._identifier}_Responses"),
        )
        self.defined_execution_errors = [
            parent_feature.defined_execution_errors[name]
            for name in xpath_sila(fdl_node, "sila:DefinedExecutionErrors/sila:Identifier/text()")
        ]

    @staticmethod
    def from_fdl_node(fdl_node, parent_feature: Feature) -> Command:
        if xpath_sila(fdl_node, "sila:Observable/text() = 'No'"):
            from sila2.framework.command.unobservable_command import UnobservableCommand

            return UnobservableCommand(fdl_node, parent_feature)
        from sila2.framework.command.observable_command import ObservableCommand

        return ObservableCommand(fdl_node, parent_feature)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List

from sila2.framework import FullyQualifiedIdentifier
from sila2.server import FeatureImplementationBase

from .silaservice_types import GetFeatureDefinition_Responses, SetServerName_Responses


class SiLAServiceBase(FeatureImplementationBase, ABC):
    @abstractmethod
    def get_ServerName(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        """
        Human readable name of the SiLA Server. The name can be set using the 'Set Server Name' command.

        :param metadata: The SiLA Client Metadata attached to the call
        :return: Human readable name of the SiLA Server. The name can be set using the 'Set Server Name' command.
        """
        pass

    @abstractmethod
    def get_ServerType(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        """
          The type of this server. It, could be, e.g., in the case of a SiLA Device the model name.
        It is specified by the implementer of the SiLA Server and MAY not be unique.

          :param metadata: The SiLA Client Metadata attached to the call
          :return: The type of this server. It, could be, e.g., in the case of a SiLA Device the model name.
        It is specified by the implementer of the SiLA Server and MAY not be unique.
        """
        pass

    @abstractmethod
    def get_ServerUUID(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        """
          Globally unique identifier that identifies a SiLA Server. The Server UUID MUST be generated once
        and remain the same for all times.

          :param metadata: The SiLA Client Metadata attached to the call
          :return: Globally unique identifier that identifies a SiLA Server. The Server UUID MUST be generated once
        and remain the same for all times.
        """
        pass

    @abstractmethod
    def get_ServerDescription(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        """
        Description of the SiLA Server. This should include the use and purpose of this SiLA Server.

        :param metadata: The SiLA Client Metadata attached to the call
        :return: Description of the SiLA Server. This should include the use and purpose of this SiLA Server.
        """
        pass

    @abstractmethod
    def get_ServerVersion(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        """
          Returns the version of the SiLA Server. A "Major" and a "Minor" version number (e.g. 1.0) MUST be provided,
        a Patch version number MAY be provided. Optionally, an arbitrary text, separated by an underscore MAY be
        appended, e.g. “3.19.373_mighty_lab_devices”.

          :param metadata: The SiLA Client Metadata attached to the call
          :return: Returns the version of the SiLA Server. A "Major" and a "Minor" version number (e.g. 1.0) MUST be provided,
        a Patch version number MAY be provided. Optionally, an arbitrary text, separated by an underscore MAY be
        appended, e.g. “3.19.373_mighty_lab_devices”.
        """
        pass

    @abstractmethod
    def get_ServerVendorURL(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> str:
        """
          Returns the URL to the website of the vendor or the website of the product of this SiLA Server.
        This URL SHOULD be accessible at all times.
        The URL is a Uniform Resource Locator as defined in RFC 1738.

          :param metadata: The SiLA Client Metadata attached to the call
          :return: Returns the URL to the website of the vendor or the website of the product of this SiLA Server.
        This URL SHOULD be accessible at all times.
        The URL is a Uniform Resource Locator as defined in RFC 1738.
        """
        pass

    @abstractmethod
    def get_ImplementedFeatures(self, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> List[str]:
        """
          Returns a list of fully qualified Feature identifiers of all implemented Features of this SiLA Server.
        This list SHOULD remain the same throughout the lifetime of the SiLA Server.

          :param metadata: The SiLA Client Metadata attached to the call
          :return: Returns a list of fully qualified Feature identifiers of all implemented Features of this SiLA Server.
        This list SHOULD remain the same throughout the lifetime of the SiLA Server.
        """
        pass

    @abstractmethod
    def GetFeatureDefinition(
        self, FeatureIdentifier: str, *, metadata: Dict[FullyQualifiedIdentifier, Any]
    ) -> GetFeatureDefinition_Responses:
        """
          Get the Feature Definition of an implemented Feature by its fully qualified Feature Identifier.
        This command has no preconditions and no further dependencies and can be called at any time.


          :param FeatureIdentifier: The fully qualified Feature identifier for which the Feature definition shall be retrieved.

          :param metadata: The SiLA Client Metadata attached to the call

          :return:

              - FeatureDefinition: The Feature definition in XML format (according to the Feature Definition Schema).


        """
        pass

    @abstractmethod
    def SetServerName(
        self, ServerName: str, *, metadata: Dict[FullyQualifiedIdentifier, Any]
    ) -> SetServerName_Responses:
        """
          Sets a human readable name to the Server Name Property.Command has no preconditions and
        no further dependencies and can be called at any time.


          :param ServerName: The human readable name to assign to the SiLA Server.

          :param metadata: The SiLA Client Metadata attached to the call

        """
        pass



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING

from lxml import etree

from sila2.framework.abc.message_mappable import MessageMappable
from sila2.framework.utils import xpath_sila

if TYPE_CHECKING:
    from sila2.framework.feature import Feature


class DataType(MessageMappable, ABC):
    @staticmethod
    def from_fdl_node(fdl_node, parent_feature: Feature, parent_namespace) -> DataType:
        sila2_framework_pb2_module = parent_feature._pb2_module.SiLAFramework__pb2
        if xpath_sila(fdl_node, "sila:Basic"):
            identifier = xpath_sila(fdl_node, "sila:Basic/text()")[0]
            if identifier == "Boolean":
                from sila2.framework.data_types.boolean import Boolean

                return Boolean(sila2_framework_pb2_module)
            if identifier == "Integer":
                from sila2.framework.data_types.integer import Integer

                return Integer(sila2_framework_pb2_module)
            if identifier == "Real":
                from sila2.framework.data_types.real import Real

                return Real(sila2_framework_pb2_module)
            if identifier == "String":
                from sila2.framework.data_types.string import String

                return String(sila2_framework_pb2_module)
            if identifier == "Binary":
                from sila2.framework.data_types.binary import Binary

                return Binary(parent_feature, sila2_framework_pb2_module)
            if identifier == "Date":
                from sila2.framework.data_types.date import Date

                return Date(sila2_framework_pb2_module)
            if identifier == "Time":
                from sila2.framework.data_types.time import Time

                return Time(sila2_framework_pb2_module)
            if identifier == "Timestamp":
                from sila2.framework.data_types.timestamp import Timestamp

                return Timestamp(sila2_framework_pb2_module)
            if identifier == "Any":
                from sila2.framework.data_types.any import Any

                return Any(sila2_framework_pb2_module)
            else:  # pragma: no cover
                raise RuntimeError(f"Unknown basic data type: {identifier}")
        if xpath_sila(fdl_node, "sila:Constrained"):
            from sila2.framework.data_types.constrained import Constrained

            return Constrained.from_fdl_node(
                xpath_sila(fdl_node, "sila:Constrained")[0],
                parent_feature,
                parent_namespace,
            )
        if xpath_sila(fdl_node, "sila:DataTypeIdentifier"):
            identifier = xpath_sila(fdl_node, "sila:DataTypeIdentifier/text()")[0]
            return parent_feature._data_type_definitions[identifier]
        if xpath_sila(fdl_node, "sila:List"):
            from sila2.framework.data_types.list import List

            return List(xpath_sila(fdl_node, "sila:List")[0], parent_feature, parent_namespace)
        if xpath_sila(fdl_node, "sila:Structure"):
            from sila2.framework.data_types.structure import Structure

            return Structure(
                xpath_sila(fdl_node, "sila:Structure")[0],
                parent_feature,
                parent_namespace,
            )
        raise RuntimeError(f"Unknown data type node: {etree.tostring(fdl_node).decode('utf-8')}")  # pragma: no cover



from sila2.framework.errors.framework_error import FrameworkError, FrameworkErrorType


class InvalidCommandExecutionUUID(FrameworkError):
    """
    Issued by a SiLA Server when a SiLA Client is trying to get or subscribe to command execution information,
    intermediate responses or responses of an observable command with an invalid command execution UUID

    Notes
    -----
    This error is raised automatically by the SDK
    """

    def __init__(self, message: str):
        super().__init__(FrameworkErrorType.INVALID_COMMAND_EXECUTION_UUID, message)



from __future__ import annotations

from typing import TYPE_CHECKING

from sila2.framework.property.property import Property

if TYPE_CHECKING:
    from sila2.framework.feature import Feature


class UnobservableProperty(Property):
    """Represents an unobservable property"""

    def __init__(self, fdl_node, parent_feature: Feature):
        super().__init__(fdl_node, parent_feature)
        self.parameter_message_type = getattr(self.parent_feature._pb2_module, f"Get_{self._identifier}_Parameters")
        self.response_message_type = getattr(self.parent_feature._pb2_module, f"Get_{self._identifier}_Responses")



from .authorizationservice_base import AuthorizationServiceBase
from .authorizationservice_client import AuthorizationServiceClient
from .authorizationservice_errors import InvalidAccessToken
from .authorizationservice_feature import AuthorizationServiceFeature

__all__ = [
    "AuthorizationServiceBase",
    "AuthorizationServiceFeature",
    "AuthorizationServiceClient",
    "InvalidAccessToken",
]



from __future__ import annotations

import os.path
import re
import sys
import tempfile
from base64 import standard_b64encode
from collections import deque
from contextlib import contextmanager
from importlib import import_module
from os.path import abspath, join
from types import ModuleType
from typing import TYPE_CHECKING, Generator, NoReturn, Tuple, Union

import grpc
from grpc_tools import protoc
from lxml import etree
from lxml.etree import XSLT, XMLParser, XMLSchema, XSLTApplyError, _Element

from sila2 import resource_dir

if TYPE_CHECKING:
    from sila2.framework.abc.sila_error import SilaError
    from sila2.framework.binary_transfer.binary_transfer_error import BinaryTransferError
    from sila2.framework.command.command import Command
    from sila2.framework.command.intermediate_response import IntermediateResponse
    from sila2.framework.command.parameter import Parameter
    from sila2.framework.command.response import Response
    from sila2.framework.data_types.data_type_definition import DataTypeDefinition
    from sila2.framework.defined_execution_error_node import DefinedExecutionErrorNode
    from sila2.framework.feature import Feature
    from sila2.framework.metadata import Metadata
    from sila2.framework.property.property import Property

    HasFullyQualifiedIdentifier = Union[
        Feature,
        Command,
        Property,
        Parameter,
        Response,
        IntermediateResponse,
        DefinedExecutionErrorNode,
        DataTypeDefinition,
        Metadata,
    ]


class FullyQualifiedIdentifierRegex:
    __originator = __category = r"[a-z][a-z\.]*"
    __identifier = r"[A-Z][a-zA-Z0-9]*"
    __major_version = r"v\d+"
    __feature_identifier = "/".join((__originator, __category, __identifier, __major_version))
    __command_identifier = "/".join((__feature_identifier, "Command", __identifier))

    FeatureIdentifier = __feature_identifier
    CommandIdentifier = __command_identifier
    CommandParameterIdentifier = "/".join((__command_identifier, "Parameter", __identifier))
    CommandResponseIdentifier = "/".join((__command_identifier, "Response", __identifier))
    IntermediateCommandResponseIdentifier = "/".join((__command_identifier, "IntermediateResponse", __identifier))
    DefinedExecutionErrorIdentifier = "/".join((__feature_identifier, "DefinedExecutionError", __identifier))
    PropertyIdentifier = "/".join((__feature_identifier, "Property", __identifier))
    TypeIdentifier = "/".join((__feature_identifier, "DataType", __identifier))
    MetadataIdentifier = "/".join((__feature_identifier, "Metadata", __identifier))


def parse_feature_definition(feature_definition: str):
    schema = XMLSchema(etree.parse(join(resource_dir, "xsd", "FeatureDefinition.xsd")))
    parser = XMLParser(schema=schema)
    return etree.fromstring(feature_definition.encode("utf-8"), parser=parser)


def xpath_sila(node, expression: str):
    """xpath with the `sila` namespace"""
    return node.xpath(expression, namespaces=dict(sila="http://www.sila-standard.org"))


def run_protoc(proto_file: str) -> Tuple[ModuleType, ModuleType]:
    path, filename = os.path.split(abspath(proto_file))
    modulename, _ = os.path.splitext(filename)

    with tempfile.TemporaryDirectory() as tmp_dir:
        protoc_args = [
            "protoc",  # protoc expects args[0] to be the program name, which is irrelevant here and will be ignored
            f"--proto_path={path}",
            f"--proto_path={abspath(join(resource_dir, 'proto'))}",
            f"--python_out={tmp_dir}",
            f"--grpc_python_out={tmp_dir}",
            proto_file,
        ]
        if filename != "SiLAFramework.proto":
            protoc_args.append(abspath(join(resource_dir, "proto", "SiLAFramework.proto")))
        if protoc.main(protoc_args) != 0:
            raise RuntimeError(f"Failed to compile proto file {proto_file}")

        with temporarily_add_to_path(tmp_dir):
            pb2_module = import_module(f"{modulename}_pb2")
            grpc_module = import_module(f"{modulename}_pb2_grpc")

        del sys.modules[pb2_module.__name__]
        del sys.modules[grpc_module.__name__]

        return pb2_module, grpc_module


def feature_definition_to_modules(fdl_node) -> Tuple[ModuleType, ModuleType]:
    xslt = XSLT(etree.parse(join(resource_dir, "xsl", "fdl2proto.xsl")))
    feature_id = xpath_sila(fdl_node, "sila:Identifier")[0].text

    try:
        proto_str = str(xslt(fdl_node))
    except XSLTApplyError as ex:
        raise ValueError(f"Invalid feature definition {feature_id}: {ex}")

    with tempfile.TemporaryDirectory() as tmp_dir:
        proto_file = join(tmp_dir, f"{feature_id}.proto")
        with open(proto_file, "w", encoding="utf-8") as proto_fp:
            proto_fp.write(proto_str)
        return run_protoc(proto_file)


@contextmanager
def temporarily_add_to_path(*paths: str):
    sys.path.extend(paths)
    try:
        yield
    finally:
        for path in paths:
            sys.path.remove(path)


def xml_node_to_normalized_string(xml_node: Union[str, _Element], remove_namespace: bool = False) -> str:
    if isinstance(xml_node, str):
        xml_node = etree.fromstring(xml_node)

    if remove_namespace:
        str_with_namespace = etree.tostring(xml_node).decode("utf-8")
        str_without_namespace = re.sub(r"^<(\w+).*?>", r"<\1>", str_with_namespace)
        node_without_namespace = etree.fromstring(str_without_namespace)
        return etree.tostring(node_without_namespace, method="c14n2", strip_text=True).decode("utf-8")

    return etree.tostring(xml_node, method="c14n2", strip_text=True).decode("utf-8")


def prettify_xml_string(xml_string: str) -> str:
    node = etree.fromstring(xml_string, parser=etree.XMLParser(remove_blank_text=True))
    return str(etree.tostring(node, pretty_print=True), "utf-8")


def raise_as_rpc_error(error: Union[SilaError, BinaryTransferError], context: grpc.ServicerContext) -> NoReturn:
    context.abort(
        grpc.StatusCode.ABORTED, details=standard_b64encode(error.to_message().SerializeToString()).decode("ascii")
    )


def consume_generator(generator: Generator):
    """
    Exhausts a generator and discards its content

    From Itertools Recipes: https://docs.python.org/3/library/itertools.html#itertools-recipes
    """
    deque(generator, maxlen=0)



from sila2.framework.errors.framework_error import FrameworkError, FrameworkErrorType


class NoMetadataAllowed(FrameworkError):
    """
    Issued by a SiLA Server when receiving a request to the SiLA Service Feature that contains SiLA Client Metadata

    Notes
    -----
    This error is raised automatically by the SDK
    """

    def __init__(self, message: str):
        super().__init__(FrameworkErrorType.NO_METADATA_ALLOWED, message)



from __future__ import annotations

from datetime import date, tzinfo
from typing import TYPE_CHECKING, NamedTuple, Optional, Type

from sila2.framework.abc.data_type import DataType
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.data_types.timezone import Timezone

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import Date as SilaDate


class SilaDateType(NamedTuple):
    date: date
    """Date"""
    timezone: tzinfo
    """Timezone"""


class Date(DataType):
    native_type = SilaDateType
    message_type: Type[SilaDate]

    def __init__(self, silaframework_pb2_module: SiLAFramework_pb2):
        self.message_type = silaframework_pb2_module.Date
        self.__timezone_field = Timezone(silaframework_pb2_module)

    def to_message(self, value: SilaDateType, toplevel_named_data_node: Optional[NamedDataNode] = None) -> SilaDate:
        d, tz = value
        if not isinstance(d, date):
            raise TypeError("Expected a date")

        return self.message_type(
            day=d.day,
            month=d.month,
            year=d.year,
            timezone=self.__timezone_field.to_message(tz),
        )

    def to_native_type(
        self, message: SilaDate, toplevel_named_data_node: Optional[NamedDataNode] = None
    ) -> SilaDateType:
        return SilaDateType(
            date(day=message.day, month=message.month, year=message.year),
            self.__timezone_field.to_native_type(message.timezone),
        )

    @staticmethod
    def from_string(value: str) -> SilaDateType:
        return SilaDateType(date.fromisoformat(value[:10]), Timezone.from_string(value[10:]))



from __future__ import annotations

from queue import Queue
from typing import Iterator, TypeVar, Union

T = TypeVar("T")


class Stream(Iterator[T]):
    """Wraps a Queue and iterates over its values"""

    __queue: Queue[Union[T, StopIteration]]
    __cancelled: bool
    is_alive: bool

    def __init__(self):
        self.__queue = Queue()
        self.__cancelled = False

    @classmethod
    def from_queue(cls, queue: Queue[T]):
        q = cls()
        q.__queue = queue
        return q

    @property
    def is_alive(self) -> bool:
        return not self.__cancelled

    def put(self, item: T) -> None:
        if not self.is_alive:
            raise RuntimeError("Cannot add item to cancelled stream")
        self.__queue.put(item)

    def cancel(self) -> None:
        if not self.is_alive:
            raise RuntimeError("Stream was already cancelled")
        self.__queue.put(StopIteration())
        self.__cancelled = True

    def __next__(self) -> T:
        item = self.__queue.get()
        if isinstance(item, StopIteration):
            self.__cancelled = True
            raise item
        return item

    def __iter__(self):
        return self



from __future__ import annotations

from typing import TYPE_CHECKING

from sila2.framework.abc.named_node import NamedNode
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier

if TYPE_CHECKING:
    from sila2.framework.feature import Feature


class DefinedExecutionErrorNode(NamedNode):
    """Represents a defined execution error"""

    _parent_feature: Feature
    fully_qualified_identifier: FullyQualifiedIdentifier
    """Fully qualified defined execution error identifier"""

    def __init__(self, fdl_node, parent_feature: Feature):
        super().__init__(fdl_node)
        self.fully_qualified_identifier = FullyQualifiedIdentifier(
            f"{parent_feature.fully_qualified_identifier}/DefinedExecutionError/{self._identifier}"
        )



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING
from typing import Any as TypingAny
from typing import NamedTuple, Optional, Type

from lxml import etree

from sila2.framework.abc.data_type import DataType
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.data_types.list import List
from sila2.framework.utils import xml_node_to_normalized_string

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import Any as SilaAny


VOID_TYPE_XML = """
<DataType>
  <Constrained>
    <DataType>
      <Basic>String</Basic>
    </DataType>
    <Constraints>
      <Length>0</Length>
    </Constraints>
  </Constrained>
</DataType>
"""


class SilaAnyType(NamedTuple):
    type_xml: str
    """
    DataType element, must adhere to
    `AnyTypeDataType.xsd <https://gitlab.com/SiLA2/sila_base/-/blob/master/schema/AnyTypeDataType.xsd>`_
    """
    value: TypingAny
    """Value of the given type"""


class Any(DataType):
    native_type = SilaAnyType
    message_type: Type[SilaAny]

    def __init__(self, silaframework_pb2_module: SiLAFramework_pb2):
        self.message_type = silaframework_pb2_module.Any

    def to_native_type(self, message: SilaAny, toplevel_named_data_node: Optional[NamedDataNode] = None) -> SilaAnyType:
        type_str = message.type
        data_type = Any.__getdata_type(type_str)
        msg = data_type.message_type.FromString(message.payload)
        if self.__is_void(type_str):
            return SilaAnyType(type_str, None)
        return SilaAnyType(type_str, data_type.to_native_type(msg, toplevel_named_data_node=toplevel_named_data_node))

    def to_message(self, value: SilaAnyType, toplevel_named_data_node: Optional[NamedDataNode] = None) -> SilaAny:
        type_str, native_payload = value
        if self.__is_void(type_str):
            native_payload = ""
        data_type = Any.__getdata_type(type_str)
        binary_payload = data_type.to_message(
            native_payload, toplevel_named_data_node=toplevel_named_data_node
        ).SerializeToString()

        return self.message_type(type=xml_node_to_normalized_string(type_str), payload=binary_payload)

    @staticmethod
    def __is_void(type_xml_str: str) -> bool:
        type_node = etree.fromstring(type_xml_str)
        return xml_node_to_normalized_string(type_node) == xml_node_to_normalized_string(VOID_TYPE_XML)

    @staticmethod
    def __getdata_type(type_xml_str: str) -> DataType:
        from sila2.framework.feature import Feature

        fdl_str = Any.__build_feature_xml(type_xml_str)
        data_type = Feature(fdl_str)._data_type_definitions["Any"]

        if isinstance(data_type, List):
            raise NotImplementedError("Any with List is not yet implemented")

        return data_type

    @staticmethod
    def __build_feature_xml(type_xml_str: str) -> str:
        return f"""<?xml version="1.0" encoding="utf-8" ?>
        <Feature SiLA2Version="1.0" FeatureVersion="1.0" Originator="org.silastandard"
                 Category="tests"
                 xmlns="http://www.sila-standard.org"
                 xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                 xsi:schemaLocation="http://www.sila-standard.org https://gitlab.com/SiLA2/sila_base/raw/master/schema/FeatureDefinition.xsd">
            <Identifier>AnyFeature{str(uuid.uuid4()).replace("-", "")}</Identifier>
            <DisplayName>Any</DisplayName>
            <Description>Dummy feature for the SiLA2 Any type</Description>
            <DataTypeDefinition>
                <Identifier>Any</Identifier>
                <DisplayName>Any</DisplayName>
                <Description>Any</Description>
                {type_xml_str}
            </DataTypeDefinition>
        </Feature>"""



"""
This class does not do anything useful at runtime. Its only purpose is to provide type annotations.
Since sphinx does not support .pyi files (yet?), this is a .py file.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:

    from typing import Iterable, Optional

    from authorizationproviderservice_types import Verify_Responses

    from sila2.client import ClientMetadataInstance


class AuthorizationProviderServiceClient:
    """
    This Feature provides SiLA Servers with the ability to check a given access token.
    """

    def Verify(
        self,
        AccessToken: str,
        RequestedServer: str,
        RequestedFeature: str,
        *,
        metadata: Optional[Iterable[ClientMetadataInstance]] = None,
    ) -> Verify_Responses:
        """
        Verifies that a given token is valid for the requested server.
        """
        ...



from sila2.framework.errors.framework_error import FrameworkError, FrameworkErrorType


class InvalidMetadata(FrameworkError):
    """
    Issued by a SiLA Server if a SiLA Client did not send the required SiLA Client Metadata,
    or when the received Metadata was invalid

    Notes
    -----
    This error is raised automatically by the SDK
    """

    def __init__(self, message: str):
        super().__init__(FrameworkErrorType.INVALID_METADATA, message)



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar, Union

from sila2.framework.abc.constraint import Constraint

if TYPE_CHECKING:
    from sila2.framework.abc.data_type import DataType
    from sila2.framework.feature import Feature


T = TypeVar("T", bound=Union[str, bytes])


class Length(Constraint[T]):
    length: int

    def __init__(self, length: int):
        if length < 0:
            raise ValueError("Length cannot be negative")
        if length > 2**63 - 1:
            raise ValueError("Length cannot be greater than 2^63 - 1")
        self.length = length

    def validate(self, value: T) -> bool:
        return len(value) == self.length

    @classmethod
    def from_fdl_node(cls, fdl_node, parent_feature: Feature, base_type: DataType) -> Length:
        return cls(int(fdl_node.text))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.length})"



from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from typing import Any, Callable, TypeVar

from grpc import RpcError
from grpc._channel import _MultiThreadedRendezvous

from sila2.client.subscription import Subscription

T = TypeVar("T")


class GrpcStreamSubscription(Subscription[T]):
    def __init__(
        self, wrapped_stream: _MultiThreadedRendezvous, converter_func: Callable[[Any], T], executor: ThreadPoolExecutor
    ) -> None:
        super().__init__()
        self.__wrapped_stream = wrapped_stream
        self.__converter_func = converter_func
        self.__executor = executor
        self.__queue = Queue()
        self.__cancelled = False

        def looped_func():
            while True:
                try:
                    new_item = self.__converter_func(next(self.__wrapped_stream))
                except (RpcError, StopIteration):
                    break

                self.__queue.put(new_item)
                for callback in self.callbacks:
                    self.__executor.submit(lambda: callback(new_item))

            self.__queue.put(StopIteration())

        executor.submit(looped_func)

    def __next__(self) -> T:
        item = self.__queue.get()
        if isinstance(item, StopIteration):
            raise item
        return item

    def cancel(self):
        self.__wrapped_stream.cancel()
        self.__queue.put(StopIteration())
        self.__cancelled = True

    @property
    def is_cancelled(self) -> bool:
        return self.__cancelled



import os
import sys
from os.path import dirname, join

if sys.version_info.minor == 7:
    import importlib_metadata as metadata
else:
    from importlib import metadata

__version__ = metadata.version(__name__)


# has to be executed before protobuf is imported
# https://github.com/protocolbuffers/protobuf/issues/3002#issuecomment-325459597
os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"

# has to be executed before grpc is imported
# https://github.com/grpc/grpc/issues/13873#issuecomment-372409083
os.environ["GRPC_ENABLE_FORK_SUPPORT"] = "false"

resource_dir = join(dirname(__file__), "resources")


__all__ = ["__version__", "resource_dir"]



import itertools
import warnings
from typing import Dict, Optional, Type
from uuid import UUID

from cryptography import x509
from zeroconf import ServiceBrowser, ServiceInfo, ServiceListener, Zeroconf

from sila2.client.sila_client import SilaClient


class SilaServiceListener(ServiceListener):
    services: Dict[str, SilaClient]

    def __init__(
        self,
        parent_browser: ServiceBrowser,
        *,
        insecure: bool = False,
        root_certs: Optional[bytes] = None,
        private_key: Optional[bytes] = None,
        cert_chain: Optional[bytes] = None,
        client_cls: Type[SilaClient] = SilaClient,
    ):
        self.parent_browser = parent_browser
        self.services = {}
        self.insecure = insecure
        self.root_certs = root_certs
        self.private_key = private_key
        self.cert_chain = cert_chain
        self.client_cls = client_cls

    def add_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        info = zc.get_service_info(type_, name)
        self.__add_client(info)

    def update_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        info = zc.get_service_info(type_, name)
        self.__add_client(info)

    def remove_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        if name in self.services:
            self.services.pop(name)

    def __add_client(self, service_info: Optional[ServiceInfo] = None) -> None:
        if service_info is None:  # happens sometimes
            return

        service_name = service_info.name
        service_uuid = UUID(service_name.split(".")[0])
        ip_address = service_info.parsed_addresses()[0]
        port = service_info.port

        try:
            if self.cert_chain is not None or self.private_key is not None or self.root_certs is not None:
                self.services[service_name] = self.client_cls(
                    ip_address,
                    port,
                    cert_chain=self.cert_chain,
                    private_key=self.private_key,
                    root_certs=self.root_certs,
                )
            elif self.insecure:
                self.services[service_name] = self.client_cls(ip_address, port, insecure=True)
            else:
                ca = self.__get_ca_from_service_info(service_info)
                if ca is not None:
                    self.services[service_name] = self.client_cls(ip_address, port, root_certs=ca)
                    warnings.warn(
                        RuntimeWarning(
                            f"Connected to server at {ip_address}:{port} with UUID {service_uuid} "
                            f"via untrusted certificate"
                        )
                    )
                else:
                    warnings.warn(
                        RuntimeWarning(
                            f"Detected server at {ip_address}:{port} with UUID {service_uuid}, "
                            f"but found no way to connect to it "
                            f"(insecure connections are disabled, no certificate information privided)"
                        )
                    )
        except Exception as ex:
            warnings.warn(
                RuntimeWarning(
                    f"Failed to connect to server at {ip_address}:{port} with UUID {service_uuid}: "
                    f"{ex.__class__.__name__} - {ex}"
                )
            )

    @staticmethod
    def __get_ca_from_service_info(service_info: ServiceInfo) -> Optional[bytes]:
        if b"ca0" not in service_info.properties:
            return None

        ca = b""
        for i in itertools.count():
            key = f"ca{i}".encode("ascii")
            if key in service_info.properties:
                ca += service_info.properties[key]
            else:
                break
        if ca == b"":
            return
        x509.load_pem_x509_certificate(ca)  # check if a full CA was found
        return ca



from __future__ import annotations

from typing import TYPE_CHECKING

from sila2.framework.abc.sila_error import SilaError

if TYPE_CHECKING:
    from sila2.client.sila_client import SilaClient
    from sila2.pb2_stubs.SiLAFramework_pb2 import SiLAError
    from sila2.pb2_stubs.SiLAFramework_pb2 import UndefinedExecutionError as SilaUndefinedExecutionError


class UndefinedExecutionError(SilaError):
    """
    Any error issued by a SiLA Server during command execution, property access, or related to SiLA Client Metadata,
    that is not a :py:class:`sila2.framework.DefinedExecutionError` (i.e. not specified in the feature definition)

    Notes
    -----
    When any exception occurs on the server-side that is no :py:class:`sila2.framework.DefinedExecutionError`
    or another SiLA error, the SDK will automatically raise it to the SiLA Client as an instance of this class
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)

    @classmethod
    def from_exception(cls, exception: Exception) -> UndefinedExecutionError:
        return cls(f"{exception.__class__.__name__} - {exception}")

    def to_message(self) -> SiLAError:
        return self._pb2_module.SiLAError(
            undefinedExecutionError=self._pb2_module.UndefinedExecutionError(message=self.message)
        )

    @classmethod
    def from_message(cls, message: SilaUndefinedExecutionError, client: SilaClient) -> UndefinedExecutionError:
        return cls(message.message)



from typing import Optional

from sila2.framework.abc.composite_message_mappable import CompositeMessageMappable
from sila2.framework.command.command import Command
from sila2.framework.command.intermediate_response import IntermediateResponse
from sila2.framework.utils import xpath_sila


class ObservableCommand(Command):
    """Represents an observable command"""

    intermediate_responses: Optional[CompositeMessageMappable[IntermediateResponse]]

    def __init__(self, fdl_node, parent_feature):
        super().__init__(fdl_node, parent_feature)
        intermediate_response_nodes = xpath_sila(fdl_node, "sila:IntermediateResponse")
        if intermediate_response_nodes:
            self.intermediate_responses = CompositeMessageMappable(
                [IntermediateResponse(node, self) for node in intermediate_response_nodes],
                getattr(self.parent_feature._pb2_module, f"{self._identifier}_IntermediateResponses"),
            )
        else:
            self.intermediate_responses = None



from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable

from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier


class MetadataInterceptor(ABC):
    def __init__(self, affected_metadata: Iterable[FullyQualifiedIdentifier]) -> None:
        """
        Intercepts SiLA Client Metadata

        Parameters
        ----------
        affected_metadata
            Fully qualified identifiers of SiLA Client Metadata to intercept

        Notes
        -----
        When a SiLA Client sends SiLA Client Metadata along with a command execution or property access request,
        the ``intercept()`` method of all affected metadata interceptors are called.
        """
        self.affected_metadata = frozenset(affected_metadata)

    @abstractmethod
    def intercept(
        self, parameters: Any, metadata: Dict[FullyQualifiedIdentifier, Any], target_call: FullyQualifiedIdentifier
    ) -> None:
        """
        Called by the SiLA Server when receiving SiLA Client Metadata that is affected by this interceptor

        Parameters
        ----------
        parameters
            Original call parameters as a named tuple
        metadata
            All SiLA Client Metadata received along with the call
        target_call
            Fully qualified identifier of the requested command or property
        """
        pass



from __future__ import annotations

import re
from datetime import timedelta, timezone, tzinfo
from typing import TYPE_CHECKING, Optional, Type

from sila2.framework.abc.message_mappable import MessageMappable
from sila2.framework.abc.named_data_node import NamedDataNode

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import Timezone as SilaTimezone


class Timezone(MessageMappable):
    native_type = tzinfo
    message_type: Type[SilaTimezone]

    def __init__(self, silaframework_pb2_module: SiLAFramework_pb2):
        self.message_type = silaframework_pb2_module.Timezone

    def to_message(self, tz: tzinfo, toplevel_named_data_node: Optional[NamedDataNode] = None) -> SilaTimezone:
        offset_timedelta = tz.utcoffset(None)

        offset_seconds = offset_timedelta.total_seconds()
        if int(offset_seconds) % 60 != 0:
            raise ValueError("SiLA2 does not support seconds in Timezone")

        offset_hours, offset_minutes = divmod(offset_seconds // 60, 60)

        return self.message_type(hours=int(offset_hours), minutes=int(offset_minutes))

    def to_native_type(
        self, message: SilaTimezone, toplevel_named_data_node: Optional[NamedDataNode] = None
    ) -> timezone:
        return timezone(timedelta(hours=message.hours, minutes=message.minutes))

    @staticmethod
    def from_string(value: str) -> tzinfo:
        if value == "Z":
            return timezone(timedelta(hours=0, minutes=0))
        if not re.match("^[+-][0-9]{2}:[0-9]{2}$", value):
            raise ValueError(f"Invalid timezone format: '{value}'. Must be 'Z' or like '+HH:MM' or '-HH:MM'")

        td = timedelta(hours=int(value[1:3]), minutes=int(value[-2:]))
        sign = int(value[0] + "1")
        td = td * sign

        if abs(td) > timedelta(hours=14):
            raise ValueError(f"Timezone UTC offset must be between -14:00 and +14:00, was {td}")

        return timezone(td)



from __future__ import annotations

import logging
from queue import Queue
from threading import Thread
from typing import TYPE_CHECKING, Callable, Generic, List, Optional, TypeVar

from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.server.observables.stream import Stream

if TYPE_CHECKING:
    from grpc import ServicerContext

T = TypeVar("T")


class SubscriptionManagerThread(Thread, Generic[T]):
    __producer_stream: Stream[T]
    __subscriber_streams: List[Stream[T]]
    __converter_func: Callable
    __last_item: Optional[T]

    def __init__(self, origin_identifier: FullyQualifiedIdentifier, producer_queue: Queue[T], converter_func: Callable):
        super().__init__(name=f"{self.__class__.__name__}-{origin_identifier}")
        self.__producer_stream = Stream.from_queue(producer_queue)
        self.__subscriber_streams = []
        self.__converter_func = converter_func
        self.__last_item = None

        _, _, feature_id, _, _, target_id = origin_identifier.split("/")
        self.logger = logging.getLogger(f"{self.__class__.__name__}[{feature_id}.{target_id}]")

    def add_subscription(self, context: ServicerContext) -> Stream[T]:
        s = Stream()

        # send last item so clients receive the current status immediately
        if self.__last_item is not None:
            s.put(self.__last_item)

        context.add_callback(lambda: self.cancel_subscription(s))

        self.__subscriber_streams.append(s)
        return s

    def cancel_subscription(self, subscription: Stream[T]):
        if subscription.is_alive:
            subscription.cancel()
        self.__subscriber_streams.remove(subscription)

    def run(self):
        for item in self.__producer_stream:
            self.logger.info(f"Received item, forwarding to all subscribers: {item}")
            self.__last_item = self.__converter_func(item)
            for subscriber in self.__subscriber_streams:
                subscriber.put(self.__last_item)

        self.__cancel_all_subscriptions()

    def __cancel_all_subscriptions(self) -> None:
        for s in self.__subscriber_streams:
            s.cancel()

    def cancel_producer(self):
        self.__producer_stream.cancel()



from __future__ import annotations

from base64 import standard_b64decode
from enum import Enum
from typing import TYPE_CHECKING

import grpc
from grpc import RpcError

from sila2.framework.abc.binary_transfer_handler import pb2_module as binary_transfer_pb2_module

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLABinaryTransfer_pb2
    from sila2.pb2_stubs.SiLABinaryTransfer_pb2 import BinaryTransferError as SilaBinaryTransferError


class BinaryTransferError(Exception):
    _pb2_module: SiLABinaryTransfer_pb2 = binary_transfer_pb2_module
    error_type: BinaryTransferErrorType
    message: str

    def __init__(self, error_type: BinaryTransferErrorType, message: str):
        self.error_type = error_type
        self.message = message
        super().__init__(message)

    def to_message(self) -> SilaBinaryTransferError:
        error_type = getattr(self._pb2_module.BinaryTransferError, self.error_type.name)
        return self._pb2_module.BinaryTransferError(errorType=error_type, message=self.message)

    @staticmethod
    def is_binary_transfer_error(exception: Exception) -> bool:
        if not isinstance(exception, RpcError):
            return False
        if not exception.code() == grpc.StatusCode.ABORTED:
            return False
        try:
            BinaryTransferError._pb2_module.BinaryTransferError.FromString(standard_b64decode(exception.details()))
        except:  # noqa: E722
            return False
        return True

    @classmethod
    def from_rpc_error(cls, rpc_error: RpcError):
        if not cls.is_binary_transfer_error(rpc_error):
            raise ValueError("Error is no BinaryTransferError")

        bin_err = cls._pb2_module.BinaryTransferError.FromString(standard_b64decode(rpc_error.details()))
        if bin_err.errorType == bin_err.BINARY_DOWNLOAD_FAILED:
            from sila2.framework.binary_transfer.binary_download_failed import BinaryDownloadFailed

            return BinaryDownloadFailed(bin_err.message)
        if bin_err.errorType == bin_err.BINARY_UPLOAD_FAILED:
            from sila2.framework.binary_transfer.binary_upload_failed import BinaryUploadFailed

            return BinaryUploadFailed(bin_err.message)
        if bin_err.errorType == bin_err.INVALID_BINARY_TRANSFER_UUID:
            from sila2.framework.binary_transfer.invalid_binary_transfer_uuid import InvalidBinaryTransferUUID

            return InvalidBinaryTransferUUID(bin_err.message)

        raise NotImplementedError(f"BinaryTransferType type not supported: {bin_err.errorType}")  # should not happen


class BinaryTransferErrorType(Enum):
    INVALID_BINARY_TRANSFER_UUID = 0
    BINARY_UPLOAD_FAILED = 1
    BINARY_DOWNLOAD_FAILED = 2



from __future__ import annotations

import math
import operator
from typing import TYPE_CHECKING, Any, List, Union

from sila2.framework.abc.constraint import Constraint
from sila2.framework.constraints.comparison_constraint import ComparisonConstraint
from sila2.framework.data_types.date import Date
from sila2.framework.data_types.integer import Integer
from sila2.framework.data_types.real import Real
from sila2.framework.data_types.string import String
from sila2.framework.data_types.time import Time
from sila2.framework.data_types.timestamp import Timestamp
from sila2.framework.utils import xpath_sila

if TYPE_CHECKING:
    from sila2.framework.feature import Feature


class Set(Constraint):
    set_values: List[SetValue]

    def __init__(
        self, fdl_node, parent_feature: Feature, base_type: Union[String, Integer, Real, Date, Time, Timestamp]
    ):
        self.set_values = []
        for value_node in xpath_sila(fdl_node, "sila:Value"):
            self.set_values.append(SetValue.from_fdl_node(value_node, parent_feature, base_type))

    def validate(self, value: Any) -> bool:
        return any(set_value.validate(value) for set_value in self.set_values)

    @classmethod
    def from_fdl_node(
        cls, fdl_node, parent_feature: Feature, base_type: Union[String, Integer, Real, Date, Time, Timestamp]
    ) -> Set:
        return cls(fdl_node, parent_feature, base_type)

    def __repr__(self) -> str:
        if isinstance(self.set_values[0].base_type, Real):
            values = [v.reference_value for v in self.set_values]
        elif isinstance(self.set_values[0].base_type, Integer):
            values = [int(v.reference_value) for v in self.set_values]  # integer values are parsed as floats
        else:
            values = [v.value_from_xml for v in self.set_values]
        return f"{self.__class__.__name__}({values})"


class SetValue(ComparisonConstraint):
    def __init__(self, value: str, base_type: Union[String, Integer, Real, Date, Time, Timestamp]):
        super().__init__(base_type, float_equals if isinstance(base_type, Real) else operator.eq, value)

    @classmethod
    def from_fdl_node(
        cls, fdl_node, parent_feature: Feature, base_type: Union[String, Integer, Real, Date, Time, Timestamp]
    ):
        text = fdl_node.text
        if text is None and isinstance(base_type, String):
            text = ""
        return cls(text, base_type)


def float_equals(x: float, y: float) -> bool:
    return math.isclose(x, y, rel_tol=1e-3)



"""
This class does not do anything useful at runtime. Its only purpose is to provide type annotations.
Since sphinx does not support .pyi files (yet?), this is a .py file.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:

    from typing import Iterable, List, Optional

    from silaservice_types import GetFeatureDefinition_Responses, SetServerName_Responses

    from sila2.client import ClientMetadataInstance, ClientUnobservableProperty


class SiLAServiceClient:
    """
    This Feature MUST be implemented by each SiLA Server.

    It specifies Commands and Properties to discover the Features a SiLA Server implements as well as details
    about the SiLA Server, like name, type, description, vendor and UUID.

    Any interaction described in this feature MUST not affect the behaviour of any other Feature.
    """

    ServerName: ClientUnobservableProperty[str]
    """
    Human readable name of the SiLA Server. The name can be set using the 'Set Server Name' command.
    """

    ServerType: ClientUnobservableProperty[str]
    """
    The type of this server. It, could be, e.g., in the case of a SiLA Device the model name.
      It is specified by the implementer of the SiLA Server and MAY not be unique.
    """

    ServerUUID: ClientUnobservableProperty[str]
    """
    Globally unique identifier that identifies a SiLA Server. The Server UUID MUST be generated once
      and remain the same for all times.
    """

    ServerDescription: ClientUnobservableProperty[str]
    """
    Description of the SiLA Server. This should include the use and purpose of this SiLA Server.
    """

    ServerVersion: ClientUnobservableProperty[str]
    """
    Returns the version of the SiLA Server. A "Major" and a "Minor" version number (e.g. 1.0) MUST be provided,
      a Patch version number MAY be provided. Optionally, an arbitrary text, separated by an underscore MAY be
      appended, e.g. “3.19.373_mighty_lab_devices”.
    """

    ServerVendorURL: ClientUnobservableProperty[str]
    """
    Returns the URL to the website of the vendor or the website of the product of this SiLA Server.
      This URL SHOULD be accessible at all times.
      The URL is a Uniform Resource Locator as defined in RFC 1738.
    """

    ImplementedFeatures: ClientUnobservableProperty[List[str]]
    """
    Returns a list of fully qualified Feature identifiers of all implemented Features of this SiLA Server.
      This list SHOULD remain the same throughout the lifetime of the SiLA Server.
    """

    def GetFeatureDefinition(
        self, FeatureIdentifier: str, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None
    ) -> GetFeatureDefinition_Responses:
        """
        Get the Feature Definition of an implemented Feature by its fully qualified Feature Identifier.
        This command has no preconditions and no further dependencies and can be called at any time.
        """
        ...

    def SetServerName(
        self, ServerName: str, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None
    ) -> SetServerName_Responses:
        """
        Sets a human readable name to the Server Name Property.Command has no preconditions and
        no further dependencies and can be called at any time.
        """
        ...



from __future__ import annotations

from typing import List, NamedTuple, Union

from sila2.code_generator.template_objects.basics import Import, Type
from sila2.framework.command.intermediate_response import IntermediateResponse
from sila2.framework.command.response import Response
from sila2.framework.data_types.structure import StructureElement


class CompositeType(NamedTuple):
    name: str
    fields: List[CompositeTypeField]

    @property
    def imports(self) -> List[Import]:
        ret = []
        for field in self.fields:
            ret.extend(field.imports)
        return ret


class CompositeTypeField(NamedTuple):
    name: str
    description: str
    type: Type

    @classmethod
    def from_field(cls, field: Union[Response, IntermediateResponse, StructureElement]) -> CompositeTypeField:
        return cls(field._identifier, field._description, Type.from_data_type(field.data_type))

    @property
    def imports(self) -> List[Import]:
        return self.type.imports


class TypeDefinition(NamedTuple):
    name: str
    type: Type

    @property
    def imports(self) -> List[Import]:
        return self.type.imports



"""
This class does not do anything useful at runtime. Its only purpose is to provide type annotations.
Since sphinx does not support .pyi files (yet?), this is a .py file.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:

    from typing import Iterable, List, Optional

    from authenticationservice_types import Login_Responses, Logout_Responses

    from sila2.client import ClientMetadataInstance


class AuthenticationServiceClient:
    """
    This Feature provides SiLA Clients with access tokens based on a user identification and password.
    1. the user needs to login with the Login command into the server with a user identification (=user name) and a password
    2. after verification, an Access Token with the Token Lifetime information will be generated and provided by the server.
    3. the user can log-out from the server with the Logout command - a valid Access Token is required to run this command.
    """

    def Login(
        self,
        UserIdentification: str,
        Password: str,
        RequestedServer: str,
        RequestedFeatures: List[str],
        *,
        metadata: Optional[Iterable[ClientMetadataInstance]] = None,
    ) -> Login_Responses:
        """
        Provides an access token based on user information.
        """
        ...

    def Logout(
        self, AccessToken: str, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None
    ) -> Logout_Responses:
        """
        Invalidates the given access token immediately.
        """
        ...



from __future__ import annotations

from datetime import timedelta
from queue import Queue
from typing import Generic, Optional, TypeVar

from sila2.framework import CommandExecutionInfo, CommandExecutionStatus


class ObservableCommandInstance:
    """Instance of a currently running observable command. Provides means to update the execution information."""

    def __init__(self, execution_info_queue: Queue[CommandExecutionInfo]):
        self.__progress: Optional[float] = None
        self.__estimated_remaining_time: Optional[timedelta] = None
        self.__status: CommandExecutionStatus = CommandExecutionStatus.running
        self.__info_queue: Queue[CommandExecutionInfo] = execution_info_queue

    @property
    def status(self) -> CommandExecutionStatus:
        """
        Command execution status. Clients are notified about updates.
        """
        return self.__status

    @status.setter
    def status(self, status: CommandExecutionStatus) -> None:
        if not isinstance(status, CommandExecutionStatus):
            raise TypeError(f"Expected a {CommandExecutionStatus.__class__.__name__}, got {status}")
        self.__status = status
        self.__update_execution_info()

    @property
    def progress(self) -> Optional[float]:
        """Command execution progress. Must be ``None``, or between 0 and 100. Clients are notified about updates."""
        return self.__progress

    @progress.setter
    def progress(self, progress: float) -> None:
        if not isinstance(progress, (int, float)):
            raise TypeError(f"Expected an int or float, got {progress}")
        if progress < 0 or progress > 100:
            raise ValueError("Progress must be between 0 and 100")
        self.__progress = progress
        self.__update_execution_info()

    @property
    def estimated_remaining_time(self) -> Optional[timedelta]:
        """Command execution progress. Must be ``None``, or a positive timedelta. Clients are notified about updates."""
        return self.__estimated_remaining_time

    @estimated_remaining_time.setter
    def estimated_remaining_time(self, estimated_remaining_time: timedelta) -> None:
        if not isinstance(estimated_remaining_time, timedelta):
            raise TypeError(f"Expected a datetime.timedelta, got {estimated_remaining_time}")
        if estimated_remaining_time.total_seconds() < 0:
            raise ValueError("Estimated remaining time cannot be negative")
        self.__estimated_remaining_time = estimated_remaining_time
        self.__update_execution_info()

    def __update_execution_info(self) -> None:
        if self.__status is None:
            self.__status = CommandExecutionStatus.running

        self.__info_queue.put(CommandExecutionInfo(self.__status, self.__progress, self.__estimated_remaining_time))


IntermediateResponseType = TypeVar("IntermediateResponseType")


class ObservableCommandInstanceWithIntermediateResponses(ObservableCommandInstance, Generic[IntermediateResponseType]):
    """
    Instance of a currently running observable command.
    Provides means to update the execution information and send intermediate responses.
    """

    def __init__(
        self,
        execution_info_queue: Queue[CommandExecutionInfo],
        intermediate_response_queue: Queue[IntermediateResponseType],
    ):
        super().__init__(execution_info_queue)
        self.__intermediate_response_queue: Queue[IntermediateResponseType] = intermediate_response_queue

    def send_intermediate_response(self, value: IntermediateResponseType) -> None:
        """
        Send intermediate responses to subscribing clients

        Parameters
        ----------
        value
            The intermediate responses to send
        """
        self.__intermediate_response_queue.put(value)



import time
from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from typing import Callable, TypeVar

from sila2.client.subscription import Subscription

_CANCELLATION_CHECK_INTERVAL: float = 0.1

T = TypeVar("T")


class PollingSubscription(Subscription[T]):
    def __init__(
        self,
        func: Callable[[], T],
        executor: ThreadPoolExecutor,
        polling_interval: float,
    ) -> None:
        super().__init__()
        self.__queue = Queue()
        self.__cancelled = False

        n_checks_per_loop = max(1, int(polling_interval / _CANCELLATION_CHECK_INTERVAL))
        polling_interval /= n_checks_per_loop

        def looped_func():
            last_item = object()
            while not self.__cancelled:
                new_item = func()
                if new_item == last_item:
                    continue  # only emit changes
                last_item = new_item

                self.__queue.put(new_item)
                for callback in self.callbacks:
                    executor.submit(lambda: callback(new_item))

                for _ in range(n_checks_per_loop):
                    if self.__cancelled:
                        break
                    time.sleep(polling_interval)
            self.__queue.put(StopIteration())

        executor.submit(looped_func)

    def __next__(self) -> T:
        item = self.__queue.get()
        if isinstance(item, StopIteration):
            raise item
        return item

    def cancel(self) -> None:
        self.__cancelled = True

    @property
    def is_cancelled(self) -> bool:
        return self.__cancelled



from .authorizationproviderservice_base import AuthorizationProviderServiceBase
from .authorizationproviderservice_client import AuthorizationProviderServiceClient
from .authorizationproviderservice_errors import AuthorizationFailed, InvalidAccessToken
from .authorizationproviderservice_feature import AuthorizationProviderServiceFeature
from .authorizationproviderservice_types import Verify_Responses

__all__ = [
    "AuthorizationProviderServiceBase",
    "AuthorizationProviderServiceFeature",
    "AuthorizationProviderServiceClient",
    "Verify_Responses",
    "AuthorizationFailed",
    "InvalidAccessToken",
]



from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta
from enum import Enum
from typing import TYPE_CHECKING, Optional, Type

from sila2.framework.abc.message_mappable import MessageMappable
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.command.duration import Duration
from sila2.framework.data_types.real import Real

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import ExecutionInfo as SilaExecutionInfo


class ExecutionInfo(MessageMappable):
    message_type: Type[SilaExecutionInfo]

    def __init__(self, silaframework_pb2_module: SiLAFramework_pb2):
        self.message_type = silaframework_pb2_module.ExecutionInfo
        self.__real_field = Real(silaframework_pb2_module)
        self.__duration_field = Duration(silaframework_pb2_module)
        self.native_type = CommandExecutionInfo

    def to_native_type(
        self, message: SilaExecutionInfo, toplevel_named_data_node: Optional[NamedDataNode] = None
    ) -> CommandExecutionInfo:
        status = getattr(CommandExecutionStatus, self.message_type.CommandStatus.Name(message.commandStatus))
        progress, estimated_remaining_time, updated_lifetime_of_execution = (
            None,
            None,
            None,
        )

        if message.HasField("progressInfo"):
            progress = self.__real_field.to_native_type(message.progressInfo)

        if message.HasField("estimatedRemainingTime"):
            estimated_remaining_time = self.__duration_field.to_native_type(message.estimatedRemainingTime)

        if message.HasField("updatedLifetimeOfExecution"):
            updated_lifetime_of_execution = self.__duration_field.to_native_type(message.updatedLifetimeOfExecution)

        return CommandExecutionInfo(
            status=status,
            progress=progress,
            estimated_remaining_time=estimated_remaining_time,
            updated_lifetime_of_execution=updated_lifetime_of_execution,
        )

    def to_message(
        self, value: CommandExecutionInfo, toplevel_named_data_node: Optional[NamedDataNode] = None
    ) -> SilaExecutionInfo:
        progress = None if value.progress is None else self.__real_field.to_message(value.progress)
        estimated_remaining_time = (
            None
            if value.estimated_remaining_time is None
            else self.__duration_field.to_message(value.estimated_remaining_time)
        )
        updated_lifetime_of_execution = (
            None
            if value.updated_lifetime_of_execution is None
            else self.__duration_field.to_message(value.updated_lifetime_of_execution)
        )

        return self.message_type(
            commandStatus=value.status.value,
            progressInfo=progress,
            estimatedRemainingTime=estimated_remaining_time,
            updatedLifetimeOfExecution=updated_lifetime_of_execution,
        )


@dataclass
class CommandExecutionInfo:
    status: CommandExecutionStatus
    progress: Optional[float] = None
    estimated_remaining_time: Optional[timedelta] = None
    updated_lifetime_of_execution: Optional[timedelta] = None


class CommandExecutionStatus(Enum):
    """Status of an observable command instance"""

    waiting = 0
    running = 1
    finishedSuccessfully = 2
    finishedWithError = 3



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Iterable, Optional, TypeVar

from sila2.client.polling_subscription import PollingSubscription
from sila2.client.subscription import Subscription
from sila2.client.utils import call_rpc_function
from sila2.framework.property.unobservable_property import UnobservableProperty

if TYPE_CHECKING:
    from sila2.client.client_feature import ClientFeature
    from sila2.client.client_metadata import ClientMetadataInstance

ItemType = TypeVar("ItemType")


class ClientUnobservableProperty(Generic[ItemType]):
    """Wraps an unobservable property"""

    def __init__(self, parent_feature: ClientFeature, wrapped_property: UnobservableProperty):
        self._parent_feature = parent_feature
        self._wrapped_property = wrapped_property

    def get(self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None) -> ItemType:
        """
        Get the current value

        Parameters
        ----------
        metadata
            SiLA Client Metadata to send along with the subscription request
        Returns
        -------
        value
            Current property value
        """
        param_msg = self._wrapped_property.get_parameters_message()

        response_msg = call_rpc_function(
            getattr(self._parent_feature._grpc_stub, f"Get_{self._wrapped_property._identifier}"),
            param_msg,
            metadata=metadata,
            client=self._parent_feature._parent_client,
            origin=self._wrapped_property,
        )

        return self._wrapped_property.to_native_type(response_msg)

    def subscribe_by_polling(
        self, poll_interval: float, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None
    ) -> Subscription[ItemType]:
        """
        Subscribe to value updates by repeatedly calling ``get()`` in a background thread.

        Parameters
        ----------
        poll_interval
            Time in seconds between calling ``get()``
        metadata
            SiLA Client Metadata to send along with the subscription request
        Returns
        -------
        subscription
            Property value subscription

        Notes
        -----
        When receiving the same value multiple times in a row, it is only emitted once.
        This simulates the subscription behavior of observable properties,
        which can be used to "observe any change of its value".
        """
        return PollingSubscription(
            lambda: self.get(metadata=metadata), self._parent_feature._parent_client._task_executor, poll_interval
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from sila2.framework.abc.sila_error import SilaError
from sila2.framework.command.parameter import Parameter
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier

if TYPE_CHECKING:
    from sila2.client.sila_client import SilaClient
    from sila2.pb2_stubs.SiLAFramework_pb2 import SiLAError
    from sila2.pb2_stubs.SiLAFramework_pb2 import ValidationError as SilaValidationError


class ValidationError(SilaError):
    """
    Issued by a SiLA Server if a SiLA Client sent invalid command parameters

    Notes
    -----
    This error is raised automatically by the SDK if a received parameter violates a constraint defined in the
    feature definition, or if the SDK could not interpret the parameter message sent by the SiLA Client
    """

    parameter_fully_qualified_identifier: FullyQualifiedIdentifier
    """Fully qualified identifier of the invalid parameter"""
    message: str
    """Error message"""

    def __init__(self, parameter: Parameter, message: str):
        self.parameter_fully_qualified_identifier = parameter.fully_qualified_identifier
        self.message = message
        super().__init__(
            f"Constraint validation failed for parameter {self.parameter_fully_qualified_identifier}: {message}"
        )

    def to_message(self) -> SiLAError:
        return self._pb2_module.SiLAError(
            validationError=self._pb2_module.ValidationError(
                parameter=self.parameter_fully_qualified_identifier, message=self.message
            )
        )

    @classmethod
    def from_message(cls, message: SilaValidationError, client: SilaClient) -> ValidationError:
        return cls(
            client._children_by_fully_qualified_identifier[FullyQualifiedIdentifier(message.parameter)], message.message
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Optional, Type
from uuid import UUID

from grpc import Channel

from sila2.framework.abc.binary_transfer_handler import BinaryTransferHandler
from sila2.framework.abc.binary_transfer_handler import grpc_module as binary_transfer_grpc_module
from sila2.framework.abc.binary_transfer_handler import pb2_module as binary_transfer_pb2_module
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.binary_transfer.binary_download_failed import BinaryDownloadFailed
from sila2.framework.binary_transfer.binary_transfer_error import BinaryTransferError
from sila2.framework.command.parameter import Parameter
from sila2.framework.utils import consume_generator

if TYPE_CHECKING:
    from sila2.pb2_stubs.SiLAFramework_pb2 import Binary as SilaBinary


class ClientBinaryTransferHandler(BinaryTransferHandler):
    upload_stub: binary_transfer_grpc_module.BinaryUploadStub
    download_stub: binary_transfer_grpc_module.BinaryDownloadStub
    chunk_size = 1024**2  # 1 MB
    known_binaries: Dict[UUID, bytes]

    def __init__(self, channel: Channel):
        self.upload_stub = binary_transfer_grpc_module.BinaryUploadStub(channel)
        self.download_stub = binary_transfer_grpc_module.BinaryDownloadStub(channel)
        self.known_binaries = {}

    def to_native_type(self, binary_uuid: UUID, toplevel_named_data_node: Optional[NamedDataNode] = None) -> bytes:
        """Get binary data from a server response"""
        try:
            if binary_uuid in self.known_binaries:
                return self.known_binaries[binary_uuid]

            size: int = self.download_stub.GetBinaryInfo(
                binary_transfer_pb2_module.GetBinaryInfoRequest(binaryTransferUUID=str(binary_uuid))
            ).binarySize

            n_chunks = self.__compute_chunk_count(size)
            chunk_requests = (
                binary_transfer_pb2_module.GetChunkRequest(
                    binaryTransferUUID=str(binary_uuid), offset=i * self.chunk_size, length=self.chunk_size
                )
                for i in range(n_chunks)
            )

            raw_result = bytearray(size)
            for chunk_response in self.download_stub.GetChunk(chunk_requests):
                raw_result[chunk_response.offset : chunk_response.offset + self.chunk_size] = chunk_response.payload

            result = bytes(raw_result)
            self.known_binaries[binary_uuid] = result

            # request deletion to free up server resources
            self.download_stub.DeleteBinary(
                binary_transfer_pb2_module.DeleteBinaryRequest(binaryTransferUUID=str(binary_uuid))
            )
            return result
        except Exception as ex:
            if BinaryTransferError.is_binary_transfer_error(ex):
                raise BinaryTransferError.from_rpc_error(ex)
            raise BinaryDownloadFailed(f"Exception during binary download: {ex}")

    def to_message(
        self, binary: bytes, message_type: Type[SilaBinary], toplevel_named_data_node: Parameter
    ) -> SilaBinary:
        """Upload binary data to server"""
        try:
            n_chunks = self.__compute_chunk_count(len(binary))
            binary_uuid = UUID(
                self.upload_stub.CreateBinary(
                    binary_transfer_pb2_module.CreateBinaryRequest(
                        binarySize=len(binary),
                        chunkCount=n_chunks,
                        parameterIdentifier=toplevel_named_data_node.fully_qualified_identifier,
                    )
                ).binaryTransferUUID
            )

            chunk_requests = (
                binary_transfer_pb2_module.UploadChunkRequest(
                    binaryTransferUUID=str(binary_uuid),
                    chunkIndex=i,
                    payload=binary[i * self.chunk_size : (i + 1) * self.chunk_size],
                )
                for i in range(self.__compute_chunk_count(len(binary)))
            )

            chunk_responses = self.upload_stub.UploadChunk(chunk_requests)
            # UploadChunk can be implemented lazily so that a request is only processed once its response is requested
            consume_generator(chunk_responses)

            return message_type(binaryTransferUUID=str(binary_uuid))
        except Exception as ex:
            try:
                raise BinaryTransferError.from_rpc_error(ex)
            except:  # noqa: E722
                raise BinaryDownloadFailed(f"Exception during binary upload: {ex}")

    def __compute_chunk_count(self, binary_size: int) -> int:
        return binary_size // self.chunk_size + (1 if binary_size % self.chunk_size != 0 else 0)



from __future__ import annotations

from typing import Optional

from sila2.framework.errors.defined_execution_error import DefinedExecutionError

from .authenticationservice_feature import AuthenticationServiceFeature


class AuthenticationFailed(DefinedExecutionError):
    def __init__(self, message: Optional[str] = None):
        if message is None:
            message = "The provided credentials are not valid."
        super().__init__(AuthenticationServiceFeature.defined_execution_errors["AuthenticationFailed"], message=message)


class InvalidAccessToken(DefinedExecutionError):
    def __init__(self, message: Optional[str] = None):
        if message is None:
            message = "The sent access token is not valid."
        super().__init__(AuthenticationServiceFeature.defined_execution_errors["InvalidAccessToken"], message=message)



from __future__ import annotations

import time
from datetime import datetime
from typing import List, Optional, Type, Union
from uuid import UUID

from zeroconf import ServiceBrowser, Zeroconf

from sila2.client.sila_client import SilaClient
from sila2.discovery.listener import SilaServiceListener


class SilaDiscoveryBrowser(ServiceBrowser):
    def __init__(
        self,
        *,
        root_certs: Optional[bytes] = None,
        private_key: Optional[bytes] = None,
        cert_chain: Optional[bytes] = None,
        insecure: bool = False,
        client_cls: Type[SilaClient] = SilaClient,
    ):
        """
        Browser that listens for SiLA Servers in the background

        Parameters
        ----------
        root_certs
            PEM-encoded root certificates
        private_key
            PEM-encoded private key
        cert_chain
            PEM-encoded certificate chain
        insecure
            If ``True``, no encryption will be used. Violates the SiLA 2 specification, only use for debugging purposes
        client_cls
            Class to be used for creating SiLA Client objects

        Notes
        -----
        Can be used as a context manager
        """
        self.listener = SilaServiceListener(
            self,
            insecure=insecure,
            root_certs=root_certs,
            private_key=private_key,
            cert_chain=cert_chain,
            client_cls=client_cls,
        )
        super().__init__(Zeroconf(), "_sila._tcp.local.", self.listener)

    @property
    def clients(self) -> List[SilaClient]:
        """List of SiLA Clients connected to all currently detected SiLA Servers"""
        return list(self.listener.services.values())

    def find_server(
        self, server_name: Optional[str] = None, server_uuid: Optional[Union[UUID, str]] = None, timeout: float = 0
    ) -> SilaClient:
        """
        Wait until a matching SiLA Server is found

        Parameters
        ----------
        server_name
            Only connect to SiLA Servers with this server name
        server_uuid
            Only connect to SiLA Servers with this server UUID
        timeout
            Time in seconds. If no matching server was found in this time, a ``TimeoutError`` will be raised

        Returns
        -------
        client
            SiLA Client connected to a matching SiLA Server

        Raises
        ------
        TimeoutError
            If no server was found
        """
        if timeout < 0:
            raise ValueError("Timeout must be non-negative")

        start_time = datetime.now()
        while timeout == 0 or (datetime.now() - start_time).total_seconds() < timeout:
            for client in self.clients:
                if server_name is not None and client.SiLAService.ServerName.get() != server_name:
                    continue
                if server_uuid is not None and client.SiLAService.ServerUUID.get() != str(server_uuid):
                    continue
                return client
            time.sleep(0.1)

        raise TimeoutError(f"No suitable SiLA server was found after {timeout} seconds")

    def __del__(self):
        if not self.done:
            self.cancel()

    def __enter__(self) -> SilaDiscoveryBrowser:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if not self.done:
            self.cancel()



from sila2.framework.errors.framework_error import FrameworkError, FrameworkErrorType


class CommandExecutionNotAccepted(FrameworkError):
    """
    Issued when a SiLA Server does not allow the requested command execution

    Notes
    -----
    This is NOT done automatically by the SDK
    """

    def __init__(self, message: str):
        super().__init__(FrameworkErrorType.COMMAND_EXECUTION_NOT_ACCEPTED, message)



"""
This class does not do anything useful at runtime. Its only purpose is to provide type annotations.
Since sphinx does not support .pyi files (yet?), this is a .py file.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:

    from typing import Iterable, Optional

    from lockcontroller_types import LockServer_Responses, UnlockServer_Responses

    from sila2.client import ClientMetadata, ClientMetadataInstance, ClientUnobservableProperty


class LockControllerClient:
    """
    This Feature allows a SiLA Client to lock a SiLA Server for exclusive use, preventing other SiLA Clients
    from using the SiLA Server while it is locked. To lock a SiLA Server a Lock Identifier has to be set, using the
    'LockServer' command. This Lock Identifier has to be sent along with every (lock protected)
    request to the SiLA Server in order to use its functionality.

    To send the lock identifier the SiLA Client Meta Data 'LockIdentifier' has to be used.

    When locking a SiLA Server a timeout can be specified that defines the time after which the SiLA Server will
    be automatically unlocked if no request with a valid lock identifier has been received meanwhile.
    After the timeout has expired or after explicit unlock no lock identifier has to be sent any more.
    """

    IsLocked: ClientUnobservableProperty[bool]
    """
    Returns true if the SiLA Server is currently locked or false else.

      This property MUST NOT be lock protected, so that any SiLA Client can query the current lock state
      of a SiLA Server.
    """

    LockIdentifier: ClientMetadata[str]
    """
    The lock identifier has to be sent with every (lock protected) call in order to use the functionality
      of a locked SiLA Server.
    """

    def LockServer(
        self, LockIdentifier: str, Timeout: int, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None
    ) -> LockServer_Responses:
        """
          Locks a SiLA Server for exclusive use by setting a lock identifier that has to be sent along with
        any following (lock protected) request as long as the SiLA Server is locked.
        The lock can be reset by issuing the 'Unlock Server' command.
        """
        ...

    def UnlockServer(
        self, LockIdentifier: str, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None
    ) -> UnlockServer_Responses:
        """
          Unlocks a locked SiLA Server. No lock identifier has to be sent for any following calls until
        the server is locked again via the 'Lock Server' command.
        """
        ...



from __future__ import annotations

import warnings
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Generic, NamedTuple, Optional, TypeVar
from uuid import UUID

from sila2.client.execution_info_subscription_thread import ExecutionInfoSubscriptionThread
from sila2.client.grpc_stream_subscription import GrpcStreamSubscription
from sila2.client.utils import call_rpc_function
from sila2.framework.command.execution_info import CommandExecutionInfo, CommandExecutionStatus

if TYPE_CHECKING:
    from sila2.client import SilaClient
    from sila2.client.client_observable_command import ClientObservableCommand
    from sila2.client.subscription import Subscription
    from sila2.pb2_stubs.SiLAFramework_pb2 import CommandExecutionUUID as SilaCommandExecutionUUID

ResponseType = TypeVar("ResponseType", bound=NamedTuple)
IntermediateResponseType = TypeVar("IntermediateResponseType", bound=NamedTuple)


class ClientObservableCommandInstance(Generic[ResponseType]):
    """
    Represents the execution of an observable command
    """

    _client_command: ClientObservableCommand
    execution_uuid: UUID
    __status: Optional[CommandExecutionStatus]
    __progress: Optional[float]
    estimated_remaining_time: Optional[timedelta]
    lifetime_of_execution: Optional[timedelta]
    __last_lifetime_update_timestamp: datetime
    __last_lifetime_update_duration: Optional[timedelta]
    __last_remaining_time_update_timestamp: datetime
    __last_remaining_time_update_duration: Optional[timedelta]
    __info_update_thread: ExecutionInfoSubscriptionThread

    def __init__(
        self,
        parent_client: SilaClient,
        client_command: ClientObservableCommand,
        execution_uuid: UUID,
        lifetime_of_execution: Optional[timedelta] = None,
    ):
        self._parent_client = parent_client
        self._client_command = client_command
        self.execution_uuid = execution_uuid
        self.__last_remaining_time_update_duration = None
        self.__last_remaining_time_update_timestamp = datetime.now()
        self.__last_lifetime_update_duration = None
        self.__last_lifetime_update_timestamp = datetime.now()
        self.__set_lifetime_of_execution(lifetime_of_execution)
        self.__status = None
        self.__progress = None
        self.__info_update_thread = ExecutionInfoSubscriptionThread(
            self._client_command._wrapped_command.fully_qualified_identifier,
            self.execution_uuid,
            getattr(client_command._parent_feature._grpc_stub, f"{client_command._wrapped_command._identifier}_Info")(
                self._get_execution_uuid_message()
            ),
            self,
        )
        self.__info_update_thread.start()

    @property
    def lifetime_of_execution(self) -> Optional[timedelta]:
        """
        Duration until the SiLA Server is allowed to forget about this command execution (``None``: infinite lifetime)
        """
        if self.__last_lifetime_update_duration is None:
            return None
        lifetime = datetime.now() - self.__last_lifetime_update_timestamp + self.__last_lifetime_update_duration
        if lifetime.total_seconds() < 0:
            return timedelta(0)
        return lifetime

    def __set_lifetime_of_execution(self, value: Optional[timedelta]) -> None:
        if self.lifetime_of_execution is None:
            if value is not None:
                warnings.warn(
                    f"Server shortened previously unconstrained lifetime of observable command instance "
                    f"{self.execution_uuid}. Ignoring."
                )
            return
        if value is None:
            warnings.warn("Server did not provide an updated lifetime of execution. Keeping the previous one.")
            return
        if value < self.lifetime_of_execution:
            warnings.warn(
                f"Server shortened the lifetime of observable command {self.execution_uuid} from "
                f"{self.lifetime_of_execution} to {value}. Ignoring."
            )
            return
        self.__last_lifetime_update_duration = value
        self.__last_lifetime_update_timestamp = datetime.now()

    @property
    def estimated_remaining_time(self) -> Optional[timedelta]:
        """Estimated remaining time (``None`` if the SiLA Server has not provided this information)"""
        if self.__last_remaining_time_update_duration is None:
            return None
        remaining = self.__last_remaining_time_update_duration - (
            datetime.now() - self.__last_remaining_time_update_timestamp
        )
        if remaining.total_seconds() < 0:
            return timedelta(0)
        return remaining

    @property
    def status(self) -> Optional[CommandExecutionStatus]:
        """Execution status (``None`` if the SiLA Server has not provided this information)"""
        return self.__status

    @property
    def progress(self) -> Optional[float]:
        """Progress in percent (``None`` if the SiLA Server has not provided this information)"""
        return self.__progress

    @property
    def done(self) -> bool:
        """``True`` if status is either ``finishedSuccessfully`` or ``finishedWithError``"""
        return self.status in (CommandExecutionStatus.finishedSuccessfully, CommandExecutionStatus.finishedWithError)

    def __set_estimated_remaining_time(self, value: Optional[timedelta]) -> None:
        if value is None:
            return
        self.__last_remaining_time_update_timestamp = datetime.now()
        self.__last_remaining_time_update_duration = value

    def _update(self, new_execution_info: CommandExecutionInfo) -> None:
        self.__status = new_execution_info.status
        self.__progress = new_execution_info.progress
        self.__set_estimated_remaining_time(new_execution_info.estimated_remaining_time)
        self.__set_lifetime_of_execution(new_execution_info.updated_lifetime_of_execution)

    def get_responses(self) -> ResponseType:
        """
        Request the command responses

        Returns
        -------
        responses
            Command responses as named tuple

        Raises
        ------
        CommandExecutionNotFinished
            If the command is still running
        SilaError
            If an error occurred during command execution
        """
        rpc_func = getattr(
            self._client_command._parent_feature._grpc_stub,
            f"{self._client_command._wrapped_command._identifier}_Result",
        )
        response_msg = call_rpc_function(
            rpc_func,
            self._get_execution_uuid_message(),
            metadata=None,
            client=self._client_command._parent_feature._parent_client,
            origin=self._client_command._wrapped_command,
        )
        return self._client_command._wrapped_command.responses.to_native_type(response_msg)

    def _get_execution_uuid_message(self) -> SilaCommandExecutionUUID:
        return self._client_command._parent_feature._pb2_module.SiLAFramework__pb2.CommandExecutionUUID(
            value=str(self.execution_uuid)
        )


class ClientObservableCommandInstanceWithIntermediateResponses(ClientObservableCommandInstance[ResponseType]):
    def subscribe_to_intermediate_responses(self) -> Subscription[IntermediateResponseType]:
        """
        Subscribe to intermediate responses

        Returns
        -------
        intermediate_response_subscription
            Subscription to the intermediate responses
        """
        rpc_func = getattr(
            self._client_command._parent_feature._grpc_stub,
            f"{self._client_command._wrapped_command._identifier}_Intermediate",
        )
        return GrpcStreamSubscription(
            rpc_func(self._get_execution_uuid_message()),
            self._client_command._wrapped_command.intermediate_responses.to_native_type,
            self._parent_client._task_executor,
        )

    def get_intermediate_response(self) -> IntermediateResponseType:
        """
        Request the current intermediate response item

        Returns
        -------
        intermediate_response : NamedTuple
            The current intermediate response

        Notes
        -----
        This is equivalent to subscribing to intermediate response and cancelling the subscription after receiving the
        first item
        """
        with self.subscribe_to_intermediate_responses() as sub:
            return next(sub)



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Iterable, Optional, TypeVar

from sila2.client.grpc_stream_subscription import GrpcStreamSubscription
from sila2.client.subscription import Subscription
from sila2.client.utils import call_rpc_function
from sila2.framework.property.observable_property import ObservableProperty

if TYPE_CHECKING:
    from sila2.client.client_feature import ClientFeature
    from sila2.client.client_metadata import ClientMetadataInstance

ItemType = TypeVar("ItemType")


class ClientObservableProperty(Generic[ItemType]):
    """Wraps an observable property"""

    def __init__(self, parent_feature: ClientFeature, wrapped_property: ObservableProperty):
        self._parent_feature = parent_feature
        self._wrapped_property = wrapped_property

    def subscribe(self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None) -> Subscription[ItemType]:
        """
        Subscribe to value updates

        Parameters
        ----------
        metadata
            SiLA Client Metadata to send along with the subscription request
        Returns
        -------
        subscription
            Property value subscription
        """
        param_msg = self._wrapped_property.get_parameters_message()
        rpc_func = getattr(self._parent_feature._grpc_stub, f"Subscribe_{self._wrapped_property._identifier}")
        response_stream = call_rpc_function(
            rpc_func,
            param_msg,
            metadata=metadata,
            client=self._parent_feature._parent_client,
            origin=self._wrapped_property,
        )
        return GrpcStreamSubscription(
            response_stream, self._wrapped_property.to_native_type, self._parent_feature._parent_client._task_executor
        )

    def get(self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None) -> ItemType:
        """
        Get the current value

        Parameters
        ----------
        metadata
            SiLA Client Metadata to send along with the subscription request

        Returns
        -------
        value
            Current property value

        Notes
        -----
        This is equivalent to subscribing to the property and cancelling the subscription after receiving the first
        item
        """
        with self.subscribe(metadata=metadata) as sub:
            return next(sub)



from os.path import dirname, join
from typing import List

from sila2.code_generator.code_generator_base import CodeGeneratorBase
from sila2.code_generator.feature_generator import FeatureGenerator
from sila2.features.authenticationservice import AuthenticationServiceFeature
from sila2.features.authorizationproviderservice import AuthorizationProviderServiceFeature
from sila2.features.authorizationservice import AuthorizationServiceFeature
from sila2.features.lockcontroller import LockControllerFeature
from sila2.framework import Feature
from sila2.server import default_feature_implementations


class CodeGenerator(CodeGeneratorBase):
    def generate_package(
        self,
        package_name: str,
        features: List[Feature],
        out_dir: str,
        *,
        lock_controller: bool = False,
        auth_features: bool = False,
        for_git_repo: bool = False,
    ) -> None:
        if not package_name.isidentifier():
            raise ValueError(f"Package name must be a valid Python identifier, was {package_name!r}")

        # remove duplicate features
        if lock_controller:
            features = [
                f for f in features if f.fully_qualified_identifier != LockControllerFeature.fully_qualified_identifier
            ]
        if auth_features:
            features = [
                f
                for f in features
                if f.fully_qualified_identifier
                not in (
                    AuthenticationServiceFeature.fully_qualified_identifier,
                    AuthorizationProviderServiceFeature.fully_qualified_identifier,
                    AuthorizationServiceFeature.fully_qualified_identifier,
                )
            ]

        # generate directories early to fail fast if they exist and overwriting is not permitted
        package_dir = join(out_dir, package_name)
        generated_dir = join(package_dir, "generated")
        implementations_dir = join(package_dir, "feature_implementations")

        self.generate_directory(out_dir, allow_overwrite=True)
        self.generate_directory(package_dir)
        self.generate_directory(generated_dir)
        self.generate_directory(implementations_dir)

        # generate setup files
        self._generate_pyproject_toml(package_name, out_dir)
        self._generate_setup_cfg(package_name, out_dir)

        # generate git repo files
        if for_git_repo:
            self.generate_file(
                join(out_dir, ".gitignore"), self.template_env.get_template("package/gitignore").render()
            )
            self.generate_file(
                join(out_dir, "README.md"),
                self.template_env.get_template("package/readme").render(package_name=package_name),
            )

        # generate package files
        self._generate_main(package_name, package_dir)
        self._generate_package_init(package_dir)
        self._generate_server(features, package_dir, lock_controller=lock_controller, auth_features=auth_features)
        self.generate_file(join(package_dir, "py.typed"), content="", allow_overwrite=True)

        # generate 'generated/'
        features_to_generate = features.copy()
        if lock_controller:
            features_to_generate.append(LockControllerFeature)
        if auth_features:
            features_to_generate.append(AuthenticationServiceFeature)
            features_to_generate.append(AuthorizationServiceFeature)
            features_to_generate.append(AuthorizationProviderServiceFeature)
        self.generate_generated_dir(features_to_generate, generated_dir)

        # generate 'feature_implementations/'
        self.generate_implementations(features, implementations_dir)
        if lock_controller:
            self.copy_default_implementation(LockControllerFeature, implementations_dir)
        if auth_features:
            self.copy_default_implementation(AuthenticationServiceFeature, implementations_dir)
            self.copy_default_implementation(AuthorizationServiceFeature, implementations_dir)
            self.copy_default_implementation(AuthorizationProviderServiceFeature, implementations_dir)
        self.generate_file(join(implementations_dir, "__init__.py"), "")

    def copy_default_implementation(self, feature: Feature, out_dir: str) -> None:
        feature_identifier = feature._identifier
        file_content = open(
            join(dirname(default_feature_implementations.__file__), f"{feature_identifier.lower()}_impl.py")
        ).read()
        file_content.replace(
            f"from sila2.features.{feature_identifier.lower()} import ",
            f"from ..generated.{feature_identifier.lower()} import",
        )
        self.generate_file(join(out_dir, f"{feature_identifier.lower()}_impl.py"), file_content)

    def generate_generated_dir(self, features: List[Feature], out_dir: str) -> None:
        self._generate_generated_init(out_dir)
        self._generate_client(features, out_dir)

        for feature in features:
            feature_dir = join(out_dir, feature._identifier.lower())
            self.generate_directory(feature_dir, allow_overwrite=True)
            feature_generator = FeatureGenerator(feature, overwrite=self.overwrite)
            feature_generator.generate_feature_files(feature_dir)
            self.generated_files.extend(feature_generator.generated_files)

    def generate_implementations(self, features: List[Feature], out_dir: str, *, prefix: str = "") -> None:
        for feature in features:
            feature_generator = FeatureGenerator(feature, overwrite=self.overwrite)
            feature_generator.generate_impl(out_dir, prefix=prefix)
            self.generated_files.extend(feature_generator.generated_files)

    def _generate_pyproject_toml(self, package_name: str, out_dir: str) -> None:
        self.generate_file(
            join(out_dir, "pyproject.toml"),
            self.template_env.get_template("package/pyproject_toml").render(),
        )

    def _generate_setup_cfg(self, package_name: str, out_dir: str) -> None:
        self.generate_file(
            join(out_dir, "setup.cfg"),
            self.template_env.get_template("package/setup_cfg").render(package_name=package_name),
        )

    def _generate_main(self, package_name: str, out_dir: str) -> None:
        self.generate_file(
            join(out_dir, "__main__.py"),
            self.template_env.get_template("package/main").render(package_name=package_name),
        )

    def _generate_package_init(self, out_dir: str) -> None:
        self.generate_file(join(out_dir, "__init__.py"), self.template_env.get_template("package/init").render())

    def _generate_server(
        self, features: List[Feature], out_dir: str, *, lock_controller: bool = False, auth_features: bool = False
    ) -> None:
        self.generate_file(
            join(out_dir, "server.py"),
            self.template_env.get_template("package/server").render(
                features=features,
                lock_controller=lock_controller,
                auth_features=auth_features,
            ),
        )

    def _generate_generated_init(self, out_dir: str) -> None:
        self.generate_file(
            join(out_dir, "__init__.py"), 'from .client import Client\n__all__ = ["Client"]', allow_overwrite=True
        )

    def _generate_client(self, features: List[Feature], out_dir: str) -> None:
        self.generate_file(
            join(out_dir, "client.py"), self.template_env.get_template("package/client").render(features=features)
        )



"""
This class does not do anything useful at runtime. Its only purpose is to provide type annotations.
Since sphinx does not support .pyi files (yet?), this is a .py file.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:

    from sila2.client import ClientMetadata


class AuthorizationServiceClient:
    """
    This Feature provides access control for the implementing server.

    It specifies the SiLA Client Metadata for the access token, that has been provided by the
    AuthenticationService core Feature.
    """

    AccessToken: ClientMetadata[str]
    """
    Token to be sent with every call in order to get access to the SiLA Server functionality.
    """



from __future__ import annotations

import logging
from datetime import timedelta
from typing import TYPE_CHECKING, Dict, Iterable, Optional
from uuid import UUID, uuid4

from grpc import ServicerContext

from sila2.framework.abc.binary_transfer_handler import grpc_module as binary_transfer_grpc_module
from sila2.framework.abc.binary_transfer_handler import pb2_module as binary_transfer_pb2_module
from sila2.framework.binary_transfer.binary_upload_failed import BinaryUploadFailed
from sila2.framework.binary_transfer.invalid_binary_transfer_uuid import InvalidBinaryTransferUUID
from sila2.framework.command.duration import Duration
from sila2.framework.utils import raise_as_rpc_error

if TYPE_CHECKING:
    from sila2.framework.binary_transfer.server_binary_transfer_handler import ServerBinaryTransferHandler
    from sila2.pb2_stubs import SiLABinaryTransfer_pb2


class BinaryUploadServicer(binary_transfer_grpc_module.BinaryUploadServicer):
    binaries_in_progress: Dict[UUID, Dict[int, Optional[bytes]]]
    _duration_field: Duration

    def __init__(self, parent_handler: ServerBinaryTransferHandler):
        self.parent_handler = parent_handler
        self.binaries_in_progress = {}
        self._duration_field = Duration(binary_transfer_pb2_module.SiLAFramework__pb2)
        self.__logger = logging.getLogger(self.__class__.__name__)

    def CreateBinary(self, request: SiLABinaryTransfer_pb2.CreateBinaryRequest, context: ServicerContext):
        self.__logger.debug(
            f"CreateBinary(binarySize={request.binarySize}, chunkCount={request.chunkCount}, "
            f"parameterIdentifier={request.parameterIdentifier})"
        )
        try:
            bin_id = uuid4()
            chunk_count = request.chunkCount

            self.binaries_in_progress[bin_id] = {i: None for i in range(chunk_count)}

            self.__logger.info(f"Created binary for {request.parameterIdentifier} with uuid {bin_id}")
            return binary_transfer_pb2_module.CreateBinaryResponse(
                binaryTransferUUID=str(bin_id), lifetimeOfBinary=self._duration_field.to_message(timedelta(minutes=1))
            )
        except Exception as ex:
            self.__logger.error(f"Failed to create binary: {ex}")
            raise_as_rpc_error(BinaryUploadFailed(f"Upload of large binary failed: {ex}"), context)

    def UploadChunk(
        self, request_iterator: Iterable[SiLABinaryTransfer_pb2.UploadChunkRequest], context: ServicerContext
    ):
        self.__logger.info("UploadChunk started")
        try:
            for chunk_request in request_iterator:
                self.__logger.debug(
                    f"Received UploadChunkRequest(binaryTransferUUID={chunk_request.binaryTransferUUID}, "
                    f"chunkIndex={chunk_request.chunkIndex}, payload=[{len(chunk_request.payload)} bytes])"
                )
                bin_id = UUID(chunk_request.binaryTransferUUID)
                index = chunk_request.chunkIndex
                payload = chunk_request.payload

                if bin_id in self.parent_handler.known_binaries:
                    self.__logger.warning(f"Received chunk for already completed binary {bin_id}. Ignoring")
                elif bin_id not in self.binaries_in_progress:
                    raise InvalidBinaryTransferUUID(f"Upload of large binary failed: invalid UUID {bin_id}")
                else:
                    self.binaries_in_progress[bin_id][index] = payload
                    if all(isinstance(b, bytes) for b in self.binaries_in_progress[bin_id].values()):
                        self.parent_handler.known_binaries[bin_id] = b"".join(
                            self.binaries_in_progress[bin_id][i]
                            for i in sorted(self.binaries_in_progress[bin_id].keys())
                        )
                        self.binaries_in_progress.pop(bin_id)
                        self.__logger.info(f"All chunks for {bin_id} received, upload complete")

                lifetime = timedelta(minutes=1)
                yield binary_transfer_pb2_module.UploadChunkResponse(
                    binaryTransferUUID=str(bin_id),
                    chunkIndex=index,
                    lifetimeOfBinary=self._duration_field.to_message(lifetime),
                )
                self.__logger.debug(
                    f"Sending UploadChunkResponse(binaryTransferUUID={bin_id}, chunkIndex={index}, "
                    f"lifetimeOfBinary=[{lifetime.total_seconds()} seconds])"
                )
            self.__logger.debug("UploadChunk completed")
        except InvalidBinaryTransferUUID as ex:
            self.__logger.error("UploadChunk failed: {ex}")
            raise_as_rpc_error(ex, context)
        except Exception as ex:
            self.__logger.error(f"UploadChunk failed: {ex}")
            raise_as_rpc_error(BinaryUploadFailed(f"Upload of large binary failed: {ex}"), context)

    def DeleteBinary(self, request: binary_transfer_pb2_module.DeleteBinaryRequest, context: ServicerContext):
        try:
            bin_id = UUID(request.binaryTransferUUID)

            if bin_id not in self.parent_handler.known_binaries:
                raise InvalidBinaryTransferUUID(f"Deletion of large binary failed: invalid UUID {bin_id}")

            self.parent_handler.known_binaries.pop(UUID(request.binaryTransferUUID))
            return binary_transfer_pb2_module.DeleteBinaryResponse()
        except InvalidBinaryTransferUUID as ex:
            raise_as_rpc_error(ex, context)
        except Exception as ex:
            raise_as_rpc_error(BinaryUploadFailed(f"Deletion of large binary failed: {ex}"), context)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable, Iterator, List, Tuple, TypeVar

ItemType = TypeVar("ItemType")


class Subscription(Iterator[ItemType], ABC):
    """Subscription to a stream of items provided by a SiLA Server"""

    def __init__(self):
        self.__callbacks: List[Callable[[ItemType], Any]] = []

    def add_callback(self, func: Callable[[ItemType], Any]) -> None:
        """
        Add a function to be called on every value update.
        Can only be used while this subscription is not cancelled.

        Parameters
        ----------
        func
            Function to be called on every value update.
            It will be called with the new value as positional parameter: ``func(new_value)``

        Raises
        ------
        RuntimeError
            If this subscription was cancelled

        Notes
        -----
        Can be used as a context manager:

        .. code-block:: python

            with thing.subscribe() as subscription:
                  do_something(subscription)
            # subscription is cancelled here

        """
        if self.is_cancelled:
            raise RuntimeError("Cannot add callbacks to cancelled streams")
        self.__callbacks.append(func)

    def clear_callbacks(self) -> None:
        """Deregister all currently registered callback functions"""
        self.__callbacks.clear()

    @property
    def callbacks(self) -> Tuple[Callable[[ItemType], Any], ...]:
        """All currently registered callback functions"""
        return tuple(self.__callbacks)

    @abstractmethod
    def cancel(self) -> None:
        """Cancel this subscription"""
        pass

    @property
    @abstractmethod
    def is_cancelled(self) -> bool:
        """``True`` is this subscription was cancelled"""
        pass

    def __enter__(self) -> Subscription[ItemType]:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if not self.is_cancelled:
            self.cancel()



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Iterable, Optional, TypeVar

from sila2.client.client_metadata import ClientMetadataInstance
from sila2.client.client_observable_command_instance import (
    ClientObservableCommandInstance,
    ClientObservableCommandInstanceWithIntermediateResponses,
)
from sila2.client.utils import call_rpc_function
from sila2.framework.command.command_confirmation import CommandConfirmation
from sila2.framework.command.observable_command import ObservableCommand

if TYPE_CHECKING:
    from sila2.client.client_feature import ClientFeature
    from sila2.pb2_stubs.SiLAFramework_pb2 import CommandConfirmation as SilaCommandConfirmation

ResultType = TypeVar("ResultType")
IntermediateType = TypeVar("IntermediateType")


class ClientObservableCommand(Generic[ResultType, IntermediateType]):
    """
    Wraps an observable command
    """

    def __init__(self, parent_feature: ClientFeature, wrapped_command: ObservableCommand):
        self._parent_feature: ClientFeature = parent_feature
        self._wrapped_command: ObservableCommand = wrapped_command

    def __call__(self, *args, **kwargs) -> ClientObservableCommandInstance:
        """
        Request execution of the observable command

        Parameters
        ----------
        args
            Command parameters as positional arguments
        kwargs
            Command parameters as keyword arguments
        metadata: Optional[Iterable[ClientMetadataInstance]]
            SiLA Client Metadata to send along with the request

        Returns
        -------
        instance
            Instance of the command execution
        """
        raw_metadata: Optional[Iterable[ClientMetadataInstance]] = kwargs.pop("metadata", None)
        param_msg = self._wrapped_command.parameters.to_message(*args, **kwargs)
        response_msg: SilaCommandConfirmation = call_rpc_function(
            getattr(self._parent_feature._grpc_stub, self._wrapped_command._identifier),
            param_msg,
            metadata=raw_metadata,
            client=self._parent_feature._parent_client,
            origin=self._wrapped_command,
        )
        exec_id, lifetime = CommandConfirmation(self._parent_feature._pb2_module.SiLAFramework__pb2).to_native_type(
            response_msg
        )
        if self._wrapped_command.intermediate_responses:
            return ClientObservableCommandInstanceWithIntermediateResponses(
                self._parent_feature._parent_client, self, exec_id, lifetime
            )
        return ClientObservableCommandInstance(self._parent_feature._parent_client, self, exec_id, lifetime)



from __future__ import annotations

import grpc
from grpc import RpcError


class SilaConnectionError(Exception):
    """
    Any error during SiLA communication that is not actively issued by the SiLA Server or SiLA Client

    Notes
    -----
    This error is raised by the :py:class:`sila2.client.SilaClient` class when a non-SiLA error is raised
    """

    exception: Exception
    """The exception that caused this error"""

    @classmethod
    def from_exception(cls, exception: Exception) -> SilaConnectionError:
        if isinstance(exception, RpcError):
            if exception.code() == grpc.StatusCode.UNAVAILABLE:
                message = "Failed to establish connection to the server"
            else:
                message = f"{exception.code().name} - {exception.details()}"
            message = f"{exception.code().name} - {message}"
        else:
            message = str(exception)
        err = cls(message)
        err.exception = exception
        return err



from __future__ import annotations

import datetime
from typing import List, NamedTuple
from typing import Type as TypingType

from sila2.framework.abc.data_type import DataType
from sila2.framework.data_types.any import Any as AnyDataType
from sila2.framework.data_types.any import SilaAnyType
from sila2.framework.data_types.binary import Binary
from sila2.framework.data_types.boolean import Boolean
from sila2.framework.data_types.constrained import Constrained
from sila2.framework.data_types.data_type_definition import DataTypeDefinition
from sila2.framework.data_types.date import Date, SilaDateType
from sila2.framework.data_types.integer import Integer
from sila2.framework.data_types.list import List as ListDataType
from sila2.framework.data_types.real import Real
from sila2.framework.data_types.string import String
from sila2.framework.data_types.structure import Structure
from sila2.framework.data_types.time import Time
from sila2.framework.data_types.timestamp import Timestamp


class Import(NamedTuple):
    origin: str
    target: str


class Type(NamedTuple):
    representation: str
    imports: List[Import]

    @classmethod
    def from_type(cls, type_: TypingType) -> Type:
        return cls(type_.__name__, [Import(type_.__module__, type_.__name__)] if type_.__module__ != "builtins" else [])

    @classmethod
    def from_data_type(cls, data_type: DataType):
        if isinstance(data_type, Integer):
            return cls.from_type(int)
        if isinstance(data_type, String):
            return cls.from_type(str)
        if isinstance(data_type, Boolean):
            return cls.from_type(bool)
        if isinstance(data_type, Binary):
            return cls.from_type(bytes)
        if isinstance(data_type, Real):
            return cls.from_type(float)
        if isinstance(data_type, Timestamp):
            return cls.from_type(datetime.datetime)
        if isinstance(data_type, Time):
            return cls.from_type(datetime.time)
        if isinstance(data_type, Date):
            return cls.from_type(SilaDateType)
        if isinstance(data_type, AnyDataType):
            return cls.from_type(SilaAnyType)
        if isinstance(data_type, Constrained):
            return cls.from_data_type(data_type.base_type)
        if isinstance(data_type, ListDataType):
            element_type = cls.from_data_type(data_type.element_type)
            list_type = cls("List", [Import("typing", "List")])
            return cls(f"List[{element_type.representation}]", list_type.imports + element_type.imports)
        if isinstance(data_type, DataTypeDefinition):
            return cls(data_type._identifier, [])
        if isinstance(data_type, Structure):
            return cls("Any", [Import("typing", "Any")])
        else:
            raise NotImplementedError(f"Generating data type {data_type} is not yet implemented")


class Field(NamedTuple):
    name: str
    type: Type
    docstring: str

    @property
    def imports(self) -> List[Import]:
        return self.type.imports



from __future__ import annotations

from typing import TYPE_CHECKING

from sila2.framework.abc.sila_error import SilaError
from sila2.framework.defined_execution_error_node import DefinedExecutionErrorNode
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier

if TYPE_CHECKING:
    from sila2.client.sila_client import SilaClient
    from sila2.pb2_stubs.SiLAFramework_pb2 import DefinedExecutionError as SilaDefinedExecutionError
    from sila2.pb2_stubs.SiLAFramework_pb2 import SiLAError


class DefinedExecutionError(SilaError):
    """
    Issued by a SiLA Server during command execution, property access, or related to SiLA Client Metadata,
    that is not a :py:class:`sila2.framework.DefinedExecutionError` (i.e. not specified in the feature definition).

    Notes
    -----
    The code generator generates a DefinedExecutionError subclass for each error defined in a feature.

    - Server Development: Never use this class directly. Always use the generated error classes.
    - Client usage: Client classes generated by the code generator will raise received defined execution errors as
        instance of the generated error classes. Dynamic clients (plain :py:class:`sila2.client.SilaClient` objects)
        raise instances of this class directly.
    """

    identifier: str
    """Error identifier, from the feature definition"""
    message: str
    """Error message"""
    fully_qualified_identifier: FullyQualifiedIdentifier
    """Fully qualified error identifier"""

    def __init__(self, error_node: DefinedExecutionErrorNode, message: str):
        self.message = message
        self.identifier = error_node._identifier
        self.fully_qualified_identifier = error_node.fully_qualified_identifier

        # if self is a specific subclass
        if self.__class__.__name__ == error_node._identifier:
            super().__init__(message)
        else:
            super().__init__(f"{error_node._identifier}: {message}")

    def to_message(self) -> SiLAError:
        return self._pb2_module.SiLAError(
            definedExecutionError=self._pb2_module.DefinedExecutionError(
                errorIdentifier=self.fully_qualified_identifier, message=self.message
            )
        )

    @classmethod
    def from_message(cls, message: SilaDefinedExecutionError, client: SilaClient) -> DefinedExecutionError:
        return cls(
            client._children_by_fully_qualified_identifier[FullyQualifiedIdentifier(message.errorIdentifier)],
            message.message,
        )



import re
from distutils.dir_util import copy_tree
from glob import glob
from os import makedirs
from os.path import basename, dirname, join
from shutil import copy

from setuptools import setup

base_dir = dirname(__file__)


def create_dir_with_init_file(path: str) -> None:
    makedirs(path, exist_ok=True)
    with open(join(path, "__init__.py"), "w"):
        pass


def copy_resources_from_sila_base() -> None:
    resource_dir = join(base_dir, "src", "sila2", "resources")
    makedirs(resource_dir, exist_ok=True)

    # xsd
    xsd_dir = join(resource_dir, "xsd")
    create_dir_with_init_file(xsd_dir)
    copy_tree(join(base_dir, "sila_base", "schema"), xsd_dir)

    # xsl
    xsl_dir = join(resource_dir, "xsl")
    create_dir_with_init_file(xsl_dir)
    for file in glob(join(base_dir, "sila_base", "xslt", "*.xsl")):
        copy(file, join(xsl_dir, basename(file)))

    # proto
    proto_dir = join(resource_dir, "proto")
    create_dir_with_init_file(proto_dir)
    copy(
        join(base_dir, "sila_base", "protobuf", "SiLAFramework.proto"),
        join(proto_dir, "SiLAFramework.proto"),
    )
    copy(
        join(base_dir, "sila_base", "protobuf", "SiLABinaryTransfer.proto"),
        join(proto_dir, "SiLABinaryTransfer.proto"),
    )


def prepare_readme() -> str:
    """README contains gitlab-internal links, which need to be extended for PyPI"""
    content = open(join(dirname(__file__), "README.md")).read()
    link_pattern = r"\]\(([^h][^)]+)\)"
    return re.sub(link_pattern, "](https://gitlab.com/sila2/sila_python/-/tree/master/\\1)", content)


copy_resources_from_sila_base()

setup(
    long_description=prepare_readme(),
)



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from sila2.framework.abc.constraint import Constraint
from sila2.framework.utils import FullyQualifiedIdentifierRegex

if TYPE_CHECKING:
    from sila2.framework.abc.data_type import DataType
    from sila2.framework.feature import Feature


class FullyQualifiedIdentifier(Constraint[str]):
    id_type: str

    def __init__(self, id_type: str):
        self.id_type = id_type

    def validate(self, value: str) -> bool:
        return bool(re.match(f"^{getattr(FullyQualifiedIdentifierRegex, self.id_type)}$", value))

    @classmethod
    def from_fdl_node(cls, fdl_node, parent_feature: Feature, base_type: DataType) -> FullyQualifiedIdentifier:
        return cls(fdl_node.text)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id_type!r})"



class FullyQualifiedIdentifier(str):
    """
    String class with case-insensitive comparison operations

    Examples
    --------
    >>> original_id = FullyQualifiedIdentifier("org.silastandard/core/SiLAService/v1")
    >>> lowercase_id = FullyQualifiedIdentifier("org.silastandard/core/silaservice/v1")
    >>> original_id == lowercase_id
    True
    """

    def __eq__(self, other):
        if isinstance(other, str):
            return self.lower() == other.lower()

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return f"{self.__class__.__name__}({super().__repr__()})"

    def __hash__(self):
        return super().lower().__hash__()



from __future__ import annotations

import datetime
import uuid
from concurrent.futures import Future
from queue import Queue
from typing import TYPE_CHECKING, Callable, Iterable

from google.protobuf.message import Message

from sila2.framework.command.execution_info import CommandExecutionInfo, CommandExecutionStatus, ExecutionInfo
from sila2.framework.command.observable_command import ObservableCommand
from sila2.framework.errors.command_execution_not_finished import CommandExecutionNotFinished
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.server.observables.observable_command_instance import (
    ObservableCommandInstance,
    ObservableCommandInstanceWithIntermediateResponses,
)
from sila2.server.observables.stream import Stream
from sila2.server.observables.subscription_manager_thread import SubscriptionManagerThread

if TYPE_CHECKING:
    from grpc import ServicerContext

    from sila2.pb2_stubs.SiLAFramework_pb2 import ExecutionInfo as SilaExecutionInfo
    from sila2.server.sila_server import SilaServer


class ObservableCommandManager:
    """Coordinates the execution of an observable command call"""

    def __init__(
        self,
        parent_server: SilaServer,
        wrapped_command: ObservableCommand,
        impl_func: Callable[[ObservableCommandInstance], Message],
        metadata_identifiers: Iterable[FullyQualifiedIdentifier],
    ) -> None:
        self.parent_server = parent_server
        self.command_execution_uuid = uuid.uuid4()
        self.wrapped_command = wrapped_command
        self.metadata_identifiers = list(metadata_identifiers)

        # prepare subscriptions
        self.execution_info_queue = Queue()
        self.intermediate_response_queue = Queue()
        self.info_subscription_thread = SubscriptionManagerThread(
            self.wrapped_command.fully_qualified_identifier,
            self.execution_info_queue,
            converter_func=ExecutionInfo(wrapped_command.parent_feature._pb2_module.SiLAFramework__pb2).to_message,
        )
        self.info_subscription_thread.start()

        if wrapped_command.intermediate_responses is not None:
            self.intermediate_response_subscription_thread = SubscriptionManagerThread(
                self.wrapped_command.fully_qualified_identifier,
                self.intermediate_response_queue,
                converter_func=wrapped_command.intermediate_responses.to_message,
            )
            self.intermediate_response_subscription_thread.start()
            command_instance = ObservableCommandInstanceWithIntermediateResponses(
                self.execution_info_queue, self.intermediate_response_queue
            )
        else:
            command_instance = ObservableCommandInstance(self.execution_info_queue)

        self.result_future = parent_server.child_task_executor.submit(impl_func, command_instance)
        self.result_future.add_done_callback(self.__after_execution)

    def subscribe_to_execution_infos(self, context: ServicerContext) -> Stream[SilaExecutionInfo]:
        if self.is_running():
            return self.info_subscription_thread.add_subscription(context)

        status = (
            CommandExecutionStatus.finishedSuccessfully
            if self.result_future.exception() is None
            else CommandExecutionStatus.finishedWithError
        )
        queue = Queue()
        queue.put(
            ExecutionInfo(self.wrapped_command.parent_feature._pb2_module.SiLAFramework__pb2).to_message(
                CommandExecutionInfo(status, 100, estimated_remaining_time=datetime.timedelta(0))
            )
        )
        stream = Stream.from_queue(queue)
        stream.cancel()
        return stream

    def subscribe_to_intermediate_responses(self, context: ServicerContext) -> Stream:
        stream = self.intermediate_response_subscription_thread.add_subscription(context)
        if not self.is_running():
            stream.cancel()  # else when subscribing after execution the stream will never end
        return stream

    def get_responses(self):
        if self.is_running():
            raise CommandExecutionNotFinished(
                f"Instance {self.command_execution_uuid} of command {self.wrapped_command._identifier} has not finished"
            )

        ex = self.result_future.exception()
        if ex is None:
            return self.result_future.result()
        raise ex

    def is_running(self) -> bool:
        return not self.result_future.done()

    def __after_execution(self, result_future: Future):
        ex = result_future.exception()
        status = CommandExecutionStatus.finishedSuccessfully if ex is None else CommandExecutionStatus.finishedWithError
        self.execution_info_queue.put(CommandExecutionInfo(status, 100, estimated_remaining_time=datetime.timedelta(0)))
        self.execution_info_queue.put(StopIteration())
        self.intermediate_response_queue.put(StopIteration())



from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any, Callable, List

if TYPE_CHECKING:
    from sila2.server import SilaServer


_STOP_THREAD_CHECK_PERIOD: float = 0.1


class FeatureImplementationBase:
    parent_server: SilaServer
    """SiLA Server serving this feature implementation"""

    def __init__(self, parent_server: SilaServer) -> None:
        """
        Base class for all SiLA feature implementations

        Parameters
        ----------
        parent_server
            SiLA Server that serves this feature implementation
        """
        self.parent_server = parent_server
        self.__started: bool = False
        self.__is_running: bool = False
        self.__periodic_funcs: List[Callable[[], None]] = []

    def start(self) -> None:
        """
        Called automatically when the SiLA Server is started

        Raises
        ------
        RuntimeError
            When called twice

        Notes
        -----
        When overriding this method, don't forget to call ``super().start()``.
        """
        if self.__started:
            raise RuntimeError("Cannot start feature implementation twice")
        self.__is_running = True
        self.__started = True

        for func in self.__periodic_funcs:
            self.parent_server.child_task_executor.submit(func)

    def stop(self) -> None:
        """
        Called automatically when the SiLA Server is stopped

        Raises
        ------
        RuntimeError
            When called while the server is not running

        Notes
        -----
        When overriding this method, don't forget to call ``super().stop()``.
        """
        if not self.__is_running:
            raise RuntimeError("Can only stop running servers")
        self.__is_running = False

    def run_periodically(self, func: Callable[[], Any], delay_seconds: float = _STOP_THREAD_CHECK_PERIOD) -> None:
        """
        Register a function to be called periodically while the SiLA Server is running

        Parameters
        ----------
        func
            Function to call periodically. Will be called without arguments. Returned values are ignored.
        delay_seconds
            Time to wait between calling ``func()``
        """
        n_checks_per_loop = max(1, int(delay_seconds / _STOP_THREAD_CHECK_PERIOD))
        delay_seconds /= n_checks_per_loop

        def looped_func():
            while self.__is_running:
                func()
                for _ in range(n_checks_per_loop):
                    if not self.__is_running:
                        break
                    time.sleep(delay_seconds)

        self.__periodic_funcs.append(looped_func)

        if self.__is_running:
            self.parent_server.child_task_executor.submit(looped_func)



from __future__ import annotations

import logging
from datetime import timedelta
from typing import TYPE_CHECKING, Iterable
from uuid import UUID

from grpc import ServicerContext

from sila2.framework.abc.binary_transfer_handler import grpc_module as binary_transfer_grpc_module
from sila2.framework.abc.binary_transfer_handler import pb2_module as binary_transfer_pb2_module
from sila2.framework.binary_transfer.binary_download_failed import BinaryDownloadFailed
from sila2.framework.binary_transfer.invalid_binary_transfer_uuid import InvalidBinaryTransferUUID
from sila2.framework.command.duration import Duration
from sila2.framework.utils import raise_as_rpc_error

if TYPE_CHECKING:
    from sila2.framework.binary_transfer.server_binary_transfer_handler import ServerBinaryTransferHandler


class BinaryDownloadServicer(binary_transfer_grpc_module.BinaryDownloadServicer):
    def __init__(self, parent_handler: ServerBinaryTransferHandler):
        self.parent_handler = parent_handler
        self._duration_field = Duration(binary_transfer_pb2_module.SiLAFramework__pb2)
        self.__logger = logging.getLogger(self.__class__.__name__)

    def GetBinaryInfo(self, request: binary_transfer_pb2_module.GetBinaryInfoRequest, context: ServicerContext):
        self.__logger.info(f"Received GetBinaryInfoRequest(binaryTransferUUID={request.binaryTransferUUID})")

        try:
            bin_id = UUID(request.binaryTransferUUID)

            if bin_id not in self.parent_handler.known_binaries:
                raise InvalidBinaryTransferUUID(f"Download of large binary failed: invalid UUID {bin_id}")

            bin_size = len(self.parent_handler.known_binaries[bin_id])
            response = binary_transfer_pb2_module.GetBinaryInfoResponse(
                binarySize=bin_size,
                lifetimeOfBinary=self._duration_field.to_message(timedelta(minutes=1)),
            )
            self.__logger.info(f"Found binary of {bin_size} bytes, sending response")
            return response
        except InvalidBinaryTransferUUID as ex:
            self.__logger.error(f"Failed to get binary info: {ex}")
            raise_as_rpc_error(ex, context)
        except Exception as ex:
            self.__logger.error(f"Failed to get binary info: {ex}")
            raise_as_rpc_error(BinaryDownloadFailed(f"Download of large binary failed: {ex}"), context)

    def GetChunk(
        self, request_iterator: Iterable[binary_transfer_pb2_module.GetChunkRequest], context: ServicerContext
    ):
        self.__logger.debug("Starting to send chunks")

        try:
            for chunk_request in request_iterator:
                self.__logger.debug(
                    f"Received GetChunkRequest(binaryTransferUUID={chunk_request.binaryTransferUUID}, "
                    f"offset={chunk_request.offset}, length={chunk_request.length})"
                )

                bin_id = UUID(chunk_request.binaryTransferUUID)

                if bin_id not in self.parent_handler.known_binaries:
                    raise InvalidBinaryTransferUUID(f"Download of large binary failed: invalid UUID {bin_id}")

                offset = chunk_request.offset
                length = chunk_request.length

                yield binary_transfer_pb2_module.GetChunkResponse(
                    binaryTransferUUID=str(bin_id),
                    offset=offset,
                    payload=self.parent_handler.known_binaries[bin_id][offset : offset + length],
                    lifetimeOfBinary=self._duration_field.to_message(timedelta(minutes=1)),
                )
                self.__logger.debug(f"Sent payload for uuid {bin_id} with offset={offset} and length={length}")
        except InvalidBinaryTransferUUID as ex:
            raise_as_rpc_error(ex, context)
            self.__logger.error(f"Failed to send chunk: {ex}")
        except Exception as ex:
            self.__logger.error(f"Failed to send chunk: {ex}")
            raise_as_rpc_error(BinaryDownloadFailed(f"Download of large binary failed: {ex}"), context)

    def DeleteBinary(self, request: binary_transfer_pb2_module.DeleteBinaryRequest, context: ServicerContext):
        try:
            bin_id = UUID(request.binaryTransferUUID)

            if bin_id not in self.parent_handler.known_binaries:
                raise InvalidBinaryTransferUUID(f"Download of large binary failed: invalid UUID {bin_id}")

            self.parent_handler.known_binaries.pop(UUID(request.binaryTransferUUID))
            return binary_transfer_pb2_module.DeleteBinaryResponse()
        except InvalidBinaryTransferUUID as ex:
            raise_as_rpc_error(ex, context)
        except Exception as ex:
            raise_as_rpc_error(BinaryDownloadFailed(f"Deletion of large binary failed: {ex}"), context)



from sila2.framework.errors.framework_error import FrameworkError, FrameworkErrorType


class CommandExecutionNotFinished(FrameworkError):
    """
    Issued by a SiLA Server when a SiLA Client is trying to get the command response of an observable command
    when the command has not been finished yet

    Notes
    -----
    This error is raised automatically by the SDK
    """

    def __init__(self, message: str):
        super().__init__(FrameworkErrorType.COMMAND_EXECUTION_NOT_FINISHED, message)



from datetime import datetime
from typing import Any, Dict, List, Union
from uuid import UUID

from sila2.features.authorizationservice import (
    AuthorizationServiceBase,
    AuthorizationServiceFeature,
    InvalidAccessToken,
)
from sila2.framework import Command, Feature, FullyQualifiedIdentifier, Property
from sila2.server import MetadataInterceptor, SilaServer
from sila2.server.default_feature_implementations.authenticationservice_impl import AccessToken


class AuthorizationServiceImpl(AuthorizationServiceBase):
    def __init__(self, parent_server: SilaServer):
        super().__init__(parent_server=parent_server)
        self.parent_server.add_metadata_interceptor(AuthorizationServiceInterceptor(parent_server))

    def get_calls_affected_by_AccessToken(self) -> List[Union[Feature, Command, Property, FullyQualifiedIdentifier]]:
        affected_features = []
        for feature in self.parent_server.features.values():
            if feature._identifier in [
                "SiLAService",
                "AuthorizationService",
                "AuthenticationService",
                "AuthorizationProviderService",
            ]:
                continue
            affected_features.append(feature)
        return affected_features


class AuthorizationServiceInterceptor(MetadataInterceptor):
    access_tokens: Dict[UUID, AccessToken]

    def __init__(self, parent_server: SilaServer):
        self.accesstoken_fqi = AuthorizationServiceFeature["AccessToken"].fully_qualified_identifier
        super().__init__([self.accesstoken_fqi])
        self.access_tokens = parent_server.feature_servicers["AuthenticationService"].implementation.active_tokens

    def intercept(
        self, parameters: Any, metadata: Dict[FullyQualifiedIdentifier, Any], target_call: FullyQualifiedIdentifier
    ) -> None:
        raw_token: str = metadata.pop(self.accesstoken_fqi)

        try:
            token = self.access_tokens[UUID(raw_token)]
        except (KeyError, ValueError):
            raise InvalidAccessToken

        if token.remaining_lifetime.total_seconds() <= 0:
            self.access_tokens.pop(token.token)
            raise InvalidAccessToken

        token.last_usage = datetime.now()



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING, Dict, Optional, Set, Type, TypeVar, Union
from uuid import UUID

import grpc

from sila2.client.client_feature import ClientFeature
from sila2.framework import DefinedExecutionError
from sila2.framework.defined_execution_error_node import DefinedExecutionErrorNode
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier

if TYPE_CHECKING:
    from sila2.features.silaservice import SiLAServiceClient
    from sila2.framework.utils import HasFullyQualifiedIdentifier

ClientCls = TypeVar("ClientCls", bound="SilaClient")


class SilaClient:
    SiLAService: SiLAServiceClient
    """
    This Feature MUST be implemented by each SiLA Server.

    It specifies Commands and Properties to discover the Features a SiLA Server implements as well as details
    about the SiLA Server, like name, type, description, vendor and UUID.

    Any interaction described in this feature MUST not affect the behaviour of any other Feature.
    """

    _channel: grpc.Channel
    _features: Dict[str, ClientFeature]
    _children_by_fully_qualified_identifier: Dict[FullyQualifiedIdentifier, HasFullyQualifiedIdentifier]
    _registered_defined_execution_error_classes: Dict[FullyQualifiedIdentifier, Type[DefinedExecutionError]]
    _task_executor: ThreadPoolExecutor

    _expected_features: Set[FullyQualifiedIdentifier] = {
        FullyQualifiedIdentifier("org.silastandard/core/SiLAService/v1"),
    }

    __address: str
    __port: int

    def __init__(
        self,
        address: str,
        port: int,
        *,
        root_certs: Optional[bytes] = None,
        private_key: Optional[bytes] = None,
        cert_chain: Optional[bytes] = None,
        insecure: bool = False,
    ):
        """
        SiLA Client, which is connected to a SiLA Server

        Parameters
        ----------
        address
            IP address or hostname of the SiLA Server
        port
            Port of the SiLA Server
        root_certs
            PEM-encoded root certificates
        private_key
            PEM-encoded private key
        cert_chain
            PEM-encoded certificate chain
        insecure
            If ``True``, no encryption will be used. Violates the SiLA 2 specification, only use for debugging purposes

        Raises
        ------
        RuntimeError
            If the server does not implement required features

        Warnings
        --------
        Using unencrypted communication violates the SiLA specification and should only be used for testing purposes
        """
        self.__address = address
        self.__port = port

        target = f"{self.__address}:{self.__port}"
        if any(par is not None for par in (root_certs, private_key, cert_chain)):
            if insecure:
                raise ValueError("Cannot use certificate information with insecure connections")

            credentials = grpc.ssl_channel_credentials(
                root_certificates=root_certs, private_key=private_key, certificate_chain=cert_chain
            )
            self._channel = grpc.secure_channel(target, credentials=credentials)
        elif insecure:
            self._channel = grpc.insecure_channel(target)
        else:
            raise ValueError("Either use an insecure connection, or provide certificate information")

        self._features = {}
        self._children_by_fully_qualified_identifier = {}
        self._registered_defined_execution_error_classes = {}

        # import locally to prevent circular import
        from sila2.features.silaservice import SiLAServiceFeature, UnimplementedFeature

        self.__add_feature(SiLAServiceFeature._feature_definition)
        self._register_defined_execution_error_class(
            SiLAServiceFeature.defined_execution_errors["UnimplementedFeature"], UnimplementedFeature
        )

        for feature_id in self.SiLAService.ImplementedFeatures.get():
            if feature_id == SiLAServiceFeature.fully_qualified_identifier:
                continue
            self.__add_feature(self.SiLAService.GetFeatureDefinition(feature_id).FeatureDefinition)

        found_features: Set[FullyQualifiedIdentifier] = {f.fully_qualified_identifier for f in self._features.values()}
        if not self._expected_features.issubset(found_features):
            raise RuntimeError(
                f"Server does not implement the following required features: {self._expected_features - found_features}"
            )

        self._task_executor = ThreadPoolExecutor(max_workers=100)

    @property
    def address(self) -> str:
        """SiLA Server IP address or hostname"""
        return self.__address

    @property
    def port(self) -> int:
        """SiLA Server port"""
        return self.__port

    @classmethod
    def discover(
        cls: Type[ClientCls],
        *,
        server_name: Optional[str] = None,
        server_uuid: Optional[Union[UUID, str]] = None,
        timeout: float = 0,
        root_certs: Optional[bytes] = None,
        private_key: Optional[bytes] = None,
        cert_chain: Optional[bytes] = None,
        insecure: bool = False,
    ) -> ClientCls:
        """
        Use SiLA Server Discovery to connect to a SiLA Server.
        If multiple matching servers are found, the returned client is connected to one of them.

        Parameters
        ----------
        server_name
            Only connect to SiLA Servers with this server name
        server_uuid
            Only connect to SiLA Servers with this server UUID
        timeout
            Time in seconds. If no matching server was found in this time, a ``TimeoutError`` will be raised
        root_certs
            PEM-encoded root certificates
        private_key
            PEM-encoded private key
        cert_chain
            PEM-encoded certificate chain
        insecure
            If ``True``, no encryption will be used. Violates the SiLA 2 specification, only use for debugging purposes

        Returns
        -------
        client
            SiLA Client connected to a discovered matching SiLA Server (instance of the calling class)

        Raises
        ------
        TimeoutError
            If no server was found in the given time
        RuntimeError
            If a server was found but did not implement required features

        Warnings
        --------
        Using unencrypted communication violates the SiLA specification and should only be used for testing purposes
        """
        from sila2.discovery.browser import SilaDiscoveryBrowser

        with SilaDiscoveryBrowser(
            insecure=insecure, root_certs=root_certs, private_key=private_key, cert_chain=cert_chain, client_cls=cls
        ) as browser:
            return browser.find_server(server_name=server_name, server_uuid=server_uuid, timeout=timeout)

    def __add_feature(self, feature_definition: str) -> None:
        feature = ClientFeature(feature_definition, self)
        self._children_by_fully_qualified_identifier[feature.fully_qualified_identifier] = feature
        self._children_by_fully_qualified_identifier.update(feature.children_by_fully_qualified_identifier)
        self._features[feature._identifier] = feature

        setattr(self, feature._identifier, feature)

    def _register_defined_execution_error_class(
        self, error_node: DefinedExecutionErrorNode, error_class: Type[DefinedExecutionError]
    ):
        self._registered_defined_execution_error_classes[error_node.fully_qualified_identifier] = error_class



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type

from sila2.framework.abc.data_type import DataType
from sila2.framework.abc.named_data_node import NamedDataNode

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import String as SilaString


class String(DataType):
    native_type = str
    message_type: Type[SilaString]

    def __init__(self, silaframework_pb2_module: SiLAFramework_pb2):
        self.message_type = silaframework_pb2_module.String

    def to_native_type(self, message: SilaString, toplevel_named_data_node: Optional[NamedDataNode] = None) -> str:
        return message.value

    def to_message(self, value: str, toplevel_named_data_node: Optional[NamedDataNode] = None) -> SilaString:
        if not isinstance(value, str):
            raise TypeError("Expected a str value")

        max_length = 2**21
        if len(value) > max_length:
            raise ValueError(f"String too long ({len(value)}, allowed: {max_length})")
        return self.message_type(value=value)

    @staticmethod
    def from_string(value: str) -> str:
        return value



from __future__ import annotations

from typing import Optional

from sila2.framework.errors.defined_execution_error import DefinedExecutionError

from .silaservice_feature import SiLAServiceFeature


class UnimplementedFeature(DefinedExecutionError):
    def __init__(self, message: Optional[str] = None):
        if message is None:
            message = "The Feature specified by the given Feature identifier is not implemented by the server."
        super().__init__(SiLAServiceFeature.defined_execution_errors["UnimplementedFeature"], message=message)
